import time
from types import SimpleNamespace as SN

from autowsgr.constants import literals
from autowsgr.fight.battle import BattlePlan
from autowsgr.fight.normal_fight import NormalFightPlan
from autowsgr.game.game_operation import (
    Expedition,
    RepairByBath,
    SetSupport,
    get_rewards,
)
from autowsgr.ocr.digit import get_loot_and_ship, get_resources
from autowsgr.scripts.main import start_script
from autowsgr.fight.exercise import NormalExercisePlan


class DailyOperation:
    def __init__(self, setting_path) -> None:
        self.timer = start_script(setting_path)

        self.config = SN(**self.timer.config.daily_automation)
        self.config.DEBUG = False
        self.complete_time = None        

        if self.config.auto_expedition:
            self.expedition_plan = Expedition(self.timer)

        if self.config.auto_battle:
            self.battle_plan = BattlePlan(
                self.timer, plan_path=f"battle/{self.config.battle_type}.yaml"
            )
        if self.config.auto_exercise:
            self.exercise_plan = NormalExercisePlan(self.timer, "exercise/plan_1.yaml")

        if self.config.auto_normal_fight:
            self.fight_plans = []
            self.fight_complete_times = []
            for plan in self.config.normal_fight_tasks:
                self.fight_plans.append(
                    NormalFightPlan(
                        self.timer,
                        plan_path=f"normal_fight/{plan[0]}.yaml",
                        fleet_id=plan[1],
                    )
                )
                self.fight_complete_times.append([0, plan[2]])  # 二元组， [已完成次数, 目标次数]

        self.start_time = self.last_time = time.time()

    def run(self):
        # 自动战役，直到超过次数
        if self.config.auto_battle:
            ret = literals.OPERATION_SUCCESS_FLAG
            while ret == literals.OPERATION_SUCCESS_FLAG:
                ret = self.battle_plan.run()

        # 自动开启战役支援
        if self.config.auto_set_support:
            SetSupport(self.timer, True)

        get_loot_and_ship(self.timer)  # 获取胖次掉落和船只掉落数据
        get_resources(self.timer)

        # 自动演习
        if self.config.auto_exercise:
            self.check_exercise()

        # 自动出征
        if self.config.auto_normal_fight:
            while self._has_unfinished() and self._ship_max():
                task_id = self._get_unfinished()

                plan = self.fight_plans[task_id]
                ret = plan.run()

                if ret == literals.OPERATION_SUCCESS_FLAG:
                    self.fight_complete_times[task_id][0] += 1
                elif ret == literals.DOCK_FULL_FLAG:
                    break  # 不解装则结束出征

                if time.time() - self.last_time >= 5 * 60:
                    self._expedition()
                    self._gain_bonus()
                    if self.config.auto_exercise:
                        self.check_exercise()
                    self.last_time = time.time()

        # 自动远征
        while True:
            if self.config.auto_exercise:
                self.check_exercise()
            self._bath_repair()
            self._expedition()
            self._gain_bonus()
            time.sleep(360)

    def _has_unfinished(self):
        return any(times[0] < times[1] for times in self.fight_complete_times)

    def _get_unfinished(self):
        for i, times in enumerate(self.fight_complete_times):
            if times[0] < times[1]:
                return i

    def _expedition(self):
        if self.config.auto_expedition:
            self.expedition_plan.run(True)

    def _gain_bonus(self):
        if self.config.auto_gain_bonus:
            get_rewards(self.timer)

    def _bath_repair(self):
        if self.config.auto_bath_repair:
            RepairByBath(self.timer)

    def _ship_max(self):
        if self.config.stop_maxship:
            if self.timer.got_ship_num < 500:
                self.timer.logger.info(f'已掉落船数量:{self.timer.got_ship_num}')
                return True
            else:
                return False
        else:
            return True
    
    def check_exercise(self):
        #判断在哪个时间段
        now_time = time.localtime(time.time())
        hour = now_time.tm_hour
        if 0 <= hour < 12:
            self.new_time_period = "0:00-12:00"
        elif 12 <= hour < 18:
            self.new_time_period = "12:00-18:00"
        else:
            self.new_time_period = "18:00-23:59"

        if self.new_time_period != self.complete_time:
            self.timer.logger.info("即将执行演习任务")
            self.exercise_plan.run()
            self.complete_time = self.new_time_period
        else:
            self.timer.logger.info("当前时间段演习已完成")



import threading as th
import time

from airtest.core.android.android import Android

from autowsgr.utils.api_image import convert_position, relative_to_absolute


class AndroidController:
    """安卓控制器

    用于提供底层的控制接口
    """

    def __init__(self, config, logger, dev: Android) -> None:
        self.config = config
        self.logger = logger
        self.dev = dev
        self.resolution = self.snapshot().shape[:2]
        self.resolution = self.resolution[::-1]

    def snapshot(self):
        return self.dev.snapshot(quality=99)

    def ShellCmd(self, cmd, *args, **kwargs):
        """向链接的模拟器发送 shell 命令
        Args:
            cmd (str):命令字符串
        """
        return self.dev.shell(cmd)

    def get_frontend_app(self):
        """获取前台应用的包名"""
        return self.ShellCmd("dumpsys window | grep mCurrentFocus")

    def start_background_app(self, package_name):
        self.dev.start_app(package_name)
        self.ShellCmd("input keyevent 3")

    def start_app(self, package_name):
        self.dev.start_app(package_name)

    def stop_app(self, package_name):
        self.dev.stop_app(package_name)

    def is_game_running(self):
        apps = self.ShellCmd("ps")
        return "zhanjian2" in apps

    def text(self, t):
        self.logger.debug(f"Typing:{t}")
        self.dev.text(t)

    def click(self, x, y, times=1, delay=0.5, enable_subprocess=False, *args, **kwargs):
        """点击模拟器相对坐标 (x,y).
        Args:
            x:相对横坐标  (相对 960x540 屏幕)

            y: 相对纵坐标  (相对 960x540 屏幕)

            delay:点击后延时(单位为秒)

            enable_subprocess:是否启用多线程加速

            Note:
                if 'enable_subprocess' is True,arg 'times' must be 1
        Returns:
            enable_subprocess == False:None

            enable_subprocess == True:A class threading.Thread refers to this click subprocess
        """

        if times < 1:
            raise ValueError("invalid arg 'times' " + str(times))
        if enable_subprocess and times != 1:
            raise ValueError("subprocess enabled but arg 'times' is not 1 but " + str(times))
        if x >= 960 or x < 0 or y >= 540 or y <= 0:
            raise ValueError(
                "invalid args 'x' or 'y',x should be in [0,960),y should be in [0,540)\n,but x is " + str(x) + ",y is " + str(y)
            )
        if delay < 0:
            raise ValueError("arg 'delay' should be positive or 0")
        original_x, original_y = x, y  # 保存原始坐标用来输出日志
        x, y = convert_position(x, y, self.resolution)
        if enable_subprocess == 1:
            if self.config.SHOW_ANDROID_INPUT and "not_show" not in kwargs:
                self.logger.debug("click:", time.time(), original_x, original_y)
            p = th.Thread(target=lambda: self.ShellCmd(f"input tap {str(x)} {str(y)}"))
            p.start()
            return p
        for _ in range(times):
            if self.config.SHOW_ANDROID_INPUT and "not_show" not in kwargs:
                self.logger.debug("click:", time.time(), original_x, original_y)
            self.ShellCmd(f"input tap {str(x)} {str(y)}")
            time.sleep(delay * self.config.DELAY)

    def relative_click(self, x, y, times=1, delay=0.5, enable_subprocess=False):
        x, y = relative_to_absolute((x, y), self.resolution)

        if times < 1:
            raise ValueError("invalid arg 'times' " + str(times))
        if delay < 0:
            raise ValueError("arg 'delay' should be positive or 0")
        if enable_subprocess and times != 1:
            raise ValueError("subprocess enabled but arg 'times' is not 1 but " + str(times))
        if enable_subprocess:
            p = th.Thread(target=lambda: self.ShellCmd(f"input tap {str(x)} {str(y)}"))
            p.start()
            return p

        for _ in range(times):
            if self.config.SHOW_ANDROID_INPUT:
                self.logger.debug("click:", time.time(), x, y)
            self.ShellCmd(f"input tap {str(x)} {str(y)}")
            time.sleep(delay * self.config.DELAY)

    def swipe(self, x1, y1, x2, y2, duration=0.5, delay=0.5, *args, **kwargs):
        """匀速滑动模拟器相对坐标 (x1,y1) 到 (x2,y2).
        Args:
            x1,y1,x2,y2:相对坐标 (960x540 屏幕)
            duration:滑动总时间
            delay:滑动后延时(单位为秒)
        """
        if delay < 0:
            raise ValueError("arg 'delay' should be positive or 0")
        if x1 >= 960 or x1 < 0 or y1 >= 540 or y1 <= 0:
            raise ValueError(
                "invalid args 'x1' or 'y1',x1 should be in [0,960),y1 should be in [0,540)\n,but x1 is "
                + str(x1)
                + ",y1 is "
                + str(y1)
            )
        if x2 >= 960 or x2 < 0 or y2 >= 540 or y2 <= 0:
            raise ValueError(
                "invalid args 'x2' or 'y2',x2 should be in [0,960),y2 should be in [0,540)\n,but x2 is "
                + str(x2)
                + ",y2 is "
                + str(y2)
            )
        x1, y1 = convert_position(x1, y1, self.resolution)
        x2, y2 = convert_position(x2, y2, self.resolution)
        duration = int(duration * 1000)
        input_str = f"input swipe {str(x1)} {str(y1)} {str(x2)} {str(y2)} {duration}"
        if self.config.SHOW_ANDROID_INPUT:
            self.logger.debug(input_str)
        self.ShellCmd(input_str)

        time.sleep(delay)

    def long_tap(self, x, y, duration=1, delay=0.5, *args, **kwargs):
        """长按相对坐标 (x,y)
        Args:
            x (_type_): 相对 (960x540 屏幕) 横坐标
            y (_type_): _description_
            duration (int, optional): 长按时间(秒). Defaults to 1.
            delay (float, optional): 操作后等待时间(秒). Defaults to 0.5.
        """
        if x >= 960 or x < 0 or y >= 540 or y <= 0:
            raise ValueError(
                "invalid args 'x' or 'y',x should be in [0,960),y should be in [0,540)\n,but x is " + str(x),
                +",y is " + str(y),
            )
        if delay < 0:
            raise ValueError("arg 'delay' should be positive or 0")
        if duration <= 0.2:
            raise ValueError("duration time too short,arg 'duration' should greater than 0.2")
        self.swipe(x, y, x, y, duration=duration, delay=delay, *args, **kwargs)



from os.path import dirname, join

DATA_ROOT = join(dirname(dirname(__file__)), "data")
IMG_ROOT = join(DATA_ROOT, "images")
MAP_ROOT = join(DATA_ROOT, "map")
SETTING_ROOT = join(DATA_ROOT, "settings")

BIN_ROOT = join(dirname(DATA_ROOT), "bin")
TUNNEL_ROOT = join(BIN_ROOT, "image_recognize")
ADB_ROOT = join(BIN_ROOT, "adb")



import copy
import os
import time

from autowsgr.constants import literals
from autowsgr.constants.custom_exceptions import ImageNotFoundErr
from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.constants.image_templates import IMG
from autowsgr.constants.positions import FIGHT_CONDITIONS_POSITION
from autowsgr.controller.run_timer import Timer
from autowsgr.game.game_operation import ChangeShips, MoveTeam, quick_repair
from autowsgr.game.get_game_info import detect_ship_stats, get_enemy_condition
from autowsgr.port.ship import Fleet
from autowsgr.utils.io import recursive_dict_update, yaml_to_dict
from autowsgr.utils.math_functions import CalcDis

from .common import DecisionBlock, FightInfo, FightPlan, FightResultInfo, start_march

"""
常规战决策模块/地图战斗用模板
"""

MAP_NUM = [5, 6, 4, 4, 5, 4, 5, 5, 3]  # 每一章的地图数量


class NormalFightInfo(FightInfo):
    # ==================== Unified Interface ====================
    def __init__(self, timer: Timer, chapter_id, map_id) -> None:
        super().__init__(timer)

        self.point_positions = None
        self.end_page = "map_page"
        self.map_image = IMG.identify_images["map_page"][0]
        self.chapter = chapter_id  # 章节名,战役为 'battle', 演习为 'exercise'
        self.map = map_id  # 节点名
        self.ship_position = (0, 0)
        self.node = "A"  # 常规地图战斗中,当前战斗点位的编号
        self.successor_states = {
            "proceed": {
                "yes": [
                    "fight_condition",
                    "spot_enemy_success",
                    "formation",
                    "fight_period",
                ],
                "no": ["map_page"],
            },
            "fight_condition": ["spot_enemy_success", "formation", "fight_period"],
            "spot_enemy_success": {
                "detour": [
                    "fight_condition",
                    "spot_enemy_success",
                    "formation",
                    "fight_period",
                ],
                "retreat": ["map_page"],
                "fight": ["formation", "fight_period"],
            },
            "formation": ["fight_period"],
            "fight_period": ["night", "result"],
            "night": {
                "yes": ["result"],
                "no": [["result", 10]],
            },
            "result": [
                "proceed",
                "map_page",
                "get_ship",
                "flagship_severe_damage",
            ],  # 两页战果
            "get_ship": ["proceed", "map_page", "flagship_severe_damage"],  # 捞到舰船
            "flagship_severe_damage": ["map_page"],
        }

        self.state2image = {
            "proceed": [IMG.fight_image[5], 7.5],
            "fight_condition": [IMG.fight_image[10], 22.5],
            "spot_enemy_success": [IMG.fight_image[2], 22.5],
            "formation": [IMG.fight_image[1], 22.5],
            "fight_period": [IMG.symbol_image[4], 30],
            "night": [IMG.fight_image[6], 150],
            "result": [IMG.fight_image[3], 90],
            "get_ship": [IMG.symbol_image[8], 5],
            "flagship_severe_damage": [IMG.fight_image[4], 7.5],
            "map_page": [self.map_image, 7.5],
        }

        self.after_match_delay = {
            "night": 1.75,
            "proceed": 0.5,
            "get_ship": 1,
        }

    def reset(self):
        self.fight_history.reset()
        self.last_state = "proceed"
        self.last_action = "yes"
        self.state = "proceed"  # 初始状态等同于 proceed 选择 yes

    def _before_match(self):
        # 点击加速
        if self.last_state in ["proceed", "fight_condition"] or self.last_action == "detour":
            self.timer.Android.click(250, 520, delay=0, enable_subprocess=True)

        self.timer.update_screen()

        # 在地图上走的过程中获取舰船位置
        if self.last_state == "proceed" or self.last_action == "detour":
            self._update_ship_position()
            self._update_ship_point()
            pos = self.timer.get_image_position(IMG.confirm_image[3], False, 0.8)
            if pos:
                self.timer.ConfirmOperation(delay=0.25, must_confirm=1, confidence=0.8)

    def _after_match(self):
        # 在某些State下可以记录额外信息
        if self.state == "spot_enemy_success":
            get_enemy_condition(self.timer, "fight")
        super()._after_match()

    # ======================== Functions ========================

    def _update_ship_position(self):
        """在战斗移动界面(有一个黄色小船在地图上跑)更新黄色小船的位置"""
        pos = self.timer.get_image_position(IMG.fight_image[7], 0, 0.8)
        if pos is None:
            pos = self.timer.get_image_position(IMG.fight_image[8], 0, 0.8)
        if pos is None:
            return
        self.ship_position = pos
        if self.config.SHOW_MAP_NODE:
            self.timer.logger.debug(self.ship_position)

    def _update_ship_point(self):
        """更新黄色小船(战斗地图上那个)所在的点位 (1-1A 这种,'A' 就是点位)"""

        self.node = "A"
        for i in range(26):
            ch = chr(ord("A") + i)
            if ch not in self.point_positions.keys():
                continue
            if CalcDis(self.point_positions[ch], self.ship_position) < CalcDis(
                self.point_positions[self.node], self.ship_position
            ):
                self.node = ch

        if self.config.SHOW_MAP_NODE:
            self.timer.logger.debug(self.node)

    def load_point_positions(self, map_path):
        """地图文件命名格式: [chapter]-[map].yaml"""
        self.point_positions = yaml_to_dict(os.path.join(map_path, str(self.chapter) + "-" + str(self.map) + ".yaml"))


def check_blood(blood, rule) -> bool:
    """检查血量状态是否满足前进条件
        >>>check_blood([None, 1, 1, 1, 2, -1, -1], [2, 2, 2, -1, -1, -1])

        >>>True
    Args:
        blood (list): 1-based
        rule (list): 0-based
    """
    l = max(len(blood) - 1, len(rule))
    for i in range(l):
        if blood[i + 1] == -1 or rule[i] == -1:
            continue
        if blood[i + 1] >= rule[i]:
            return False
    return True


class NormalFightPlan(FightPlan):
    """ " 常规战斗的决策模块"""

    """ 多点战斗基本模板 """

    def __init__(self, timer: Timer, plan_path, fleet_id=None, fleet=-1) -> None:
        """初始化决策模块,可以重新指定默认参数,优先级更高

        Args:
            fleet_id: 指定舰队编号, 如果为 None 则使用计划中的参数

            fleet: 舰队成员, ["", "1号位", "2号位", ...], 如果为 None 则全部不变, 为 "" 则该位置无舰船, 为 -1 则不覆盖 yaml 文件中的参数
        Raises:
            BaseException: _description_
        """

        super().__init__(timer)

        # 从配置文件加载计划
        default_args = yaml_to_dict(os.path.join(self.config.PLAN_ROOT, "default.yaml"))
        plan_args = yaml_to_dict(os.path.join(self.config.PLAN_ROOT, plan_path))

        # 从参数加载计划
        if fleet_id is not None:
            plan_args["fleet_id"] = fleet_id  # 舰队编号
        if fleet != -1:
            plan_args["fleet"] = fleet

        # 检查参数完整情况
        if "fleet_id" not in plan_args:
            self.logger.warning(f"未指定作战舰队, 默认采用第 {default_args['fleet_id']} 舰队作战")

        # 从默认参数加载
        plan_defaults = default_args["normal_fight_defaults"]
        plan_defaults.update({"node_defaults": default_args["node_defaults"]})
        args = recursive_dict_update(plan_defaults, plan_args, skip=["node_args"])
        self.__dict__.update(args)

        # 加载节点配置
        self.nodes = {}
        for node_name in self.selected_nodes:
            node_args = copy.deepcopy(self.node_defaults)
            if "node_args" in plan_args and plan_args["node_args"] is not None and node_name in plan_args["node_args"]:
                node_args = recursive_dict_update(node_args, plan_args["node_args"][node_name])
            self.nodes[node_name] = DecisionBlock(timer, node_args)
        self._load_fight_info()

    def _load_fight_info(self):
        # 信息记录器
        self.Info = NormalFightInfo(self.timer, self.chapter, self.map)
        self.Info.load_point_positions(os.path.join(MAP_ROOT, "normal"))

    def _go_map_page(self):
        """活动多点战斗必须重写该模块"""
        """ 进入选择战斗地图的页面 """
        """ 这个模块负责的工作在战斗结束后如果需要进行重复战斗, 则不会进行 """
        self.timer.goto_game_page("map_page")

    def _go_fight_prepare_page(self) -> None:
        """活动多点战斗必须重写该模块"""
        """(从当前战斗结束后跳转到的页面)进入准备战斗的页面"""
        self.timer.goto_game_page("fight_prepare_page")

    def _enter_fight(self, same_work=False, *args, **kwargs):
        """
        从任意界面进入战斗.

        :return: 进入战斗状态信息，包括['success', 'dock is full'].
        """
        if not same_work:
            self._go_map_page()
            self._change_fight_map(self.chapter, self.map)
        # try:
        assert self.timer.wait_images(self.Info.map_image) != None
        self._go_fight_prepare_page()
        MoveTeam(self.timer, self.fleet_id)
        if self.fleet is not None and same_work == False:
            ChangeShips(self.timer, self.fleet_id, self.fleet)
        self.Info.ship_stats = detect_ship_stats(self.timer)
        quick_repair(self.timer, self.repair_mode, self.Info.ship_stats)
        # TODO: 这里应该只catch network error，太宽的catch会导致其他错误被隐藏
        # except AssertionError:
        #     if "process_err" in kwargs and kwargs["process_err"] == False:
        #         raise ImageNotFoundErr("进入战斗前置界面错误")
        #     self.logger.warning("进入战斗前置界面不正确")
        #     self.timer.process_bad_network()
        #     return self._enter_fight(process_err=False)

        return start_march(self.timer)

    def _make_decision(self):  # sourcery skip: extract-duplicate-method
        state = self.update_state()
        # 进行 MapLevel 的决策
        if state == "map_page":
            self.Info.fight_history.add_event("自动回港", {"position": self.Info.node, "info": "正常"})
            return literals.FIGHT_END_FLAG

        elif state == "fight_condition":
            value = self.fight_condition
            self.timer.Android.click(*FIGHT_CONDITIONS_POSITION[value])
            self.Info.last_action = value
            self.Info.fight_history.add_event("战况选择", {"position": self.Info.node}, value)
            # self.fight_recorder.append(StageRecorder(self.Info, self.timer))
            return literals.FIGHT_CONTINUE_FLAG

        # 不在白名单之内 SL
        if self.Info.node not in self.selected_nodes:
            # 可以撤退点撤退
            if state == "spot_enemy_success":
                self.timer.Android.click(677, 492, delay=0)
                self.Info.last_action = "retreat"
                self.Info.fight_history.add_event("索敌成功", {"position": self.Info.node, "enemys": "不在预设点, 不进行索敌"}, "撤退")
                return literals.FIGHT_END_FLAG
            # 不能撤退退游戏
            elif state == "formation":
                self.Info.fight_history.add_event("阵型选择", {"position": self.Info.node}, "SL")
                return "need SL"
            elif state == "fight_period":
                self.Info.fight_history.add_event("进入战斗", {"position": self.Info.node}, "SL")
                return "need SL"

        elif state == "proceed":
            is_proceed = self.nodes[self.Info.node].proceed and check_blood(
                self.Info.ship_stats, self.nodes[self.Info.node].proceed_stop
            )

            if is_proceed:
                self.timer.Android.click(325, 350)
                self.Info.last_action = "yes"
                self.Info.fight_history.add_event(
                    "继续前进",
                    {"position": self.Info.node, "ship_stats": self.Info.ship_stats},
                    "前进",
                )
                return literals.FIGHT_CONTINUE_FLAG
            else:
                self.timer.Android.click(615, 350)
                self.Info.last_action = "no"
                self.Info.fight_history.add_event(
                    "继续前进",
                    {"position": self.Info.node, "ship_stats": self.Info.ship_stats},
                    "回港",
                )
                return literals.FIGHT_END_FLAG

        elif state == "flagship_severe_damage":
            self.timer.click_image(IMG.fight_image[4], must_click=True, delay=0.25)
            self.Info.fight_history.add_event("自动回港", {"position": self.Info.node, "info": "旗舰大破"})
            return "fight end"

        # Todo:燃油耗尽自动回港

        # 进行通用 NodeLevel 决策
        action, fight_stage = self.nodes[self.Info.node].make_decision(
            state, self.Info.last_state, self.Info.last_action, self.Info
        )
        self.Info.last_action = action
        # self.fight_recorder.append(StageRecorder(self.Info, self.timer))

        return fight_stage

    # ======================== Functions ========================
    def _get_chapter(self) -> int:
        """在出征界面获取当前章节(远征界面也可获取)

        Raises:
            TimeoutError: 无法获取当前章节

        Returns:
            int: 当前章节
        """
        for try_times in range(5):
            time.sleep(0.15 * 2**try_times)
            self.timer.update_screen()
            for i in range(1, len(IMG.chapter_image)):
                if self.timer.image_exist(IMG.chapter_image[i], 0):
                    return i

        raise TimeoutError("can't verify chapter")

    def _move_chapter(self, target_chapter, now_chapter=None):
        """移动地图章节到 target_chapter
        含错误检查

        Args:
            target_chapter (int): 目标
            now_chapter (_type_, optional): 现在的章节. Defaults to None.
        Raise:
            ImageNotFoundErr:如果没有找到章节标志或地图界面标志
        """
        if not self.timer.identify_page("map_page"):
            raise ImageNotFoundErr("not on page 'map_page' now")

        if now_chapter == target_chapter:
            return
        try:
            if now_chapter is None:
                now_chapter = self._get_chapter()
            if self.config.SHOW_CHAPTER_INFO:
                self.timer.logger.debug("NowChapter:", now_chapter)
            if now_chapter > target_chapter:
                if now_chapter - target_chapter >= 3:
                    now_chapter -= 3
                    self.timer.Android.click(95, 97, delay=0)

                elif now_chapter - target_chapter == 2:
                    now_chapter -= 2
                    self.timer.Android.click(95, 170, delay=0)

                elif now_chapter - target_chapter == 1:
                    now_chapter -= 1
                    self.timer.Android.click(95, 229, delay=0)

            else:
                if now_chapter - target_chapter <= -3:
                    now_chapter += 3
                    self.timer.Android.click(95, 485, delay=0)

                elif now_chapter - target_chapter == -2:
                    now_chapter += 2
                    self.timer.Android.click(95, 416, delay=0)

                elif now_chapter - target_chapter == -1:
                    now_chapter += 1
                    self.timer.Android.click(95, 366, delay=0)

            if not self.timer.wait_image(IMG.chapter_image[now_chapter]):
                raise ImageNotFoundErr("after 'move chapter' operation but the chapter do not move")

            time.sleep(0.15)
            self._move_chapter(target_chapter, now_chapter)
        except:
            self.logger.error(f"切换章节失败 target_chapter: {target_chapter}   now: {now_chapter}")
            if self.timer.process_bad_network("move_chapter"):
                self._move_chapter(target_chapter)
            else:
                raise ImageNotFoundErr("unknow reason can't find chapter image")

    def _verify_map(self, target_map, chapter, need_screen_shot=True, timeout=0):
        if timeout == 0:
            return self.timer.image_exist(
                IMG.normal_map_image[f"{str(chapter)}-{str(target_map)}"],
                need_screen_shot,
                confidence=0.85,
            )
        return self.timer.wait_image(
            IMG.normal_map_image[f"{str(chapter)}-{str(target_map)}"],
            confidence=0.85,
            timeout=timeout,
            gap=0.03,
        )

    def _get_map(self, chapter, need_screen_shot=True) -> int:
        """出征界面获取当前显示地图节点编号
        例如在出征界面显示的地图 2-5,则返回 5

        Returns:
            int: 节点编号
        """
        for try_times in range(5):
            time.sleep(0.15 * 2**try_times)
            if need_screen_shot:
                self.timer.update_screen()

            # 通过+-1来匹配0，1开始的序号
            for map in range(1, MAP_NUM[chapter - 1] + 1):
                if self._verify_map(map, chapter, need_screen_shot=False):
                    return map

        raise TimeoutError("can't verify map")

    def _move_map(self, target_map, chapter):
        """改变地图节点,不检查是否有该节点
        含网络错误检查
        Args:
            target_map (_type_): 目标节点

        """
        if not self.timer.identify_page("map_page"):
            raise ImageNotFoundErr("not on page 'map_page' now")

        now_map = self._get_map(chapter)
        try:
            if self.config.SHOW_CHAPTER_INFO:
                self.timer.logger.debug("now_map:", now_map)
            if target_map > now_map:
                for i in range(target_map - now_map):
                    self.timer.Android.swipe(715, 147, 552, 147, duration=0.25)
                    if not self._verify_map(now_map + (i + 1), chapter, timeout=4):
                        raise ImageNotFoundErr("after 'move map' operation but the chapter do not move")
                    time.sleep(0.15)
            else:
                for i in range(now_map - target_map):
                    self.timer.Android.swipe(552, 147, 715, 147, duration=0.25)
                    if not self._verify_map(now_map - (i + 1), chapter, timeout=4):
                        raise ImageNotFoundErr("after 'move map' operation but the chapter do not move")
                    time.sleep(0.15)
        except:
            self.logger.error(f"切换地图失败 target_map: {target_map}   now: {now_map}")
            if self.timer.process_bad_network():
                self._move_map(target_map, chapter)
            else:
                raise ImageNotFoundErr("unknown reason can't find number image" + str(target_map))

    def _change_fight_map(self, chapter, map):
        """活动多点战斗必须重写该模块"""
        """ 这个模块负责的工作在战斗结束后如果需要进行重复战斗, 则不会进行 """
        """在地图界面改变战斗地图(通常是为了出征)
        可以处理网络错误
        Args:
            chapter (int): 目标章节
            map (int): 目标节点

        Raises:
            ValueError: 不在地图界面
            ValueError: 不存在的节点
        """
        if self.timer.now_page.name != "map_page":
            raise ValueError("can't change fight map at page:", self.timer.now_page.name)
        if map - 1 not in range(MAP_NUM[chapter - 1]):
            raise ValueError(f"map {str(map)} not in the list of chapter {str(chapter)}")

        self._move_chapter(chapter)
        self._move_map(map, chapter)
        self.Info.chapter = self.chapter
        self.Info.map = self.map



ALL_SHIP_TYPES = [
    "BB",
    "BC",
    "CL",
    "CV",
    "CA",
    "CVL",
    "NAP",
    "DD",
    "SS",
    "CLT",
    "KP",
    "CG" "BM",
    "AV",
    "AADG",
    "ASDG",
    "BG",
    "CBG",
    "NO",
    "BBV",
    "CAV",
    "SC",
    "SAP",
    "KP",
]
ALL_SHIP_TYPES_CN = [
    "轻巡",
    "驱逐",
    "航母",
    "潜艇",
    "导驱",
    "航战",
    "重巡",
    "防驱",
    "战列",
    "战巡",
    "导驱",
    "导战",
    "大巡",
    "补给",
    "机场",
    "雷巡",
    "装母",
    "轻母",
    "补给",
    "重炮",
    "航巡",
    "导巡",
    "防巡",
]

CG = "CG"
KP = "KP"
BB = "BB"
BC = "BC"
CL = "CL"
CV = "CV"
CA = "CA"
CVL = "CVL"
NAP = "NAP"
DD = "DD"
SS = "SS"
CLT = "CLT"
BM = "BM"
AV = "AV"
AADG = "AADG"
ASDG = "ASDG"
BG = "BG"
CBG = "CBG"
NO = "NO"
BBV = "BBV"
CAV = "CAV"
SC = "SC"
SAP = "SAP"

CN_TYPE_TO_EN_TYPE = {
    "轻巡": CL,
    "重巡": CA,
    "雷巡": CLT,
    "航巡": CAV,
    "大巡": CBG,
    "导巡": KP,
    "防巡": CG,
    "航母": CV,
    "轻母": CVL,
    "装母": AV,
    "战列": BB,
    "航战": BBV,
    "战巡": BC,
    "重炮": BM,
    "驱逐": DD,
    "炮潜": SC,
    "潜艇": SS,
    "补给": NAP,
    "导驱": ASDG,
    "防驱": AADG,
    "导战": BG,
    "机场": "AF",
}

"""舰船类型,参考游戏中显示的字符串,特殊和陌生的命名在下面给出,路基等单位暂时无数据
    NAP:常规补给舰, 胖次期特殊点位刷六个那种,或者说除了掉胖次的之外的所有(包括我方)
    SAP:特殊补给舰, 胖次期掉胖次的蓝色补给舰
    CBG:导弹巡洋舰
    CAV:航巡
    SC:炮潜
    BG:导战
    NO:该位置不存在舰船
"""

ALL_PAGES = {
    "map_page",
    "main_page",
    "decisive_battle_entrance",
    "exercise_page",
    "battle_page",
    "expedition_page",
    "fight_prepare_page",
    "bath_page",
    "choose_repair_page",
    "backyard_page",
    "canteen_page",
    "options_page",
    "build_page",
    "destroy_page",
    "develop_page",
    "discard_page",
    "intensify_page",
    "remake_page",
    "skill_page",
    "mission_page",
    "support_set_page",
    "friend_page",
    "friend_home_page",
    "decisive_battle_entrance",
}
"""名字说明:
'fight_prepare_page':包含了快速修理,补给,综合属性等选项的界面
'friend_home_page':好友演习界面(从好友界面点进去之后的，好友的提督室)
"""

RESOURCE_NAME = {
    "oil",
    "ammo",
    "steel",
    "aluminum",
    "diamond",
    "quick_repair",
    "quick_build",
    "ship_blueprint",
    "equipment_blueprint",
}

# 船舱容量,装备容量
PORT_FACILITIES = {"ship_limit", "ship_amount", "equipment_limit", "equipment_amount"}

# 维修船坞, 建造位, 开发位

CHAPTER_NODE_COUNT = [0, 5, 6, 4, 4, 5, 4, 5, 5, 3]
# 常规图节点数



import copy
import datetime
import os
import time
from typing import Iterable, Tuple

import cv2

from autowsgr.constants.custom_exceptions import ImageNotFoundErr
from autowsgr.constants.image_templates import make_dir_templates
from autowsgr.utils.api_image import MyTemplate, convert_position, locateCenterOnImage
from autowsgr.utils.math_functions import CalcDis
from autowsgr.utils.new_logger import Logger

from .android_controller import AndroidController
from .windows_controller import WindowsController


class Emulator:
    """模拟器管理单位, 可以用于其它游戏, 对 AndroidController 的进一步封装"""

    def __init__(self, config, logger: Logger):
        # 获取设置，初始化windows控制器
        self.config = config
        self.logger = logger
        self.Windows = WindowsController(config.emulator, logger)

        # 获取额外图像数据
        self._add_extra_images()

        # 初始化android控制器
        dev = self.Windows.connect_android()
        self.Android = AndroidController(config, logger, dev)
        self.update_screen()
        self.config.resolution = self.screen.shape[:2]
        self.config.resolution = self.config.resolution[::-1]  # 转换为 （宽x高）
        self.logger.info(f"resolution:{str(self.config.resolution)}")

    # ==========初始化函数==========

    def _add_extra_images(self):
        if "EXTRA_IMAGE_ROOT" in self.config.__dict__ and self.config.EXTRA_IMAGE_ROOT is not None:
            if os.path.isdir(self.config.EXTRA_IMAGE_ROOT):
                self.images = make_dir_templates(self.config.EXTRA_IMAGE_ROOT)
                self.logger.info(f"Extra Images Loaded:{len(self.images)}")
            elif self.config.EXTRA_IMAGE_ROOT is not None:
                self.logger.warning("配置文件参数 EXTRA_IMAGE_ROOT 存在但不是合法的路径")

    # ===========命令函数===========

    def shell(self, cmd, *args, **kwargs):
        """向链接的模拟器发送 adb shell 命令"""
        return self.Android.ShellCmd(cmd)

    def get_frontend_app(self):
        """获取前台应用的包名"""
        return self.shell("dumpsys window | grep mCurrentFocus")

    def start_app(self, package_name):
        """启动应用"""
        self.Android.start_app(package_name)

    def stop_app(self, package_name):
        """停止应用"""
        self.Android.stop_app(package_name)

    def list_apps(self):
        """列出所有正在运行的应用"""
        return self.shell("ps")

    def is_running(self, app="zhanjian2"):
        """检查一个应用是否在运行

        Args:
            app (str, optional): 应用名, 默认为 "战舰少女R".

        Returns:
            bool:
        """

        return "zhanjian2" in self.list_apps()

    # ===========控制函数============

    def text(self, str):
        """输入文本

        需要焦点在输入框时才能输入
        """
        self.Android.text(str)

    def click(self, x, y):
        """点击模拟器坐标
        Args:
            x (int): 相对横坐标(标准为 960x540 大小的屏幕)
            y (int): 相对纵坐标
        Examples:
            >>> emulator=Emulator(config, logger)
            >>> emulator.click(432, 221)
        """
        self.Android.click(x, y, delay=0.1)

    def swipe(self, x0, y0, x1, y1, duration=0.5):
        """从 (x0, y0) 滑动到 (x1, y1)
        Args:
            x0: 相对横坐标(标准为 960x540 大小的屏幕)

            duration (float): 滑动时间, 单位为秒
        """
        self.Android.swipe(x0, y0, x1, y1, duration=duration, delay=0.1)

    # ===========图像函数============

    def update_screen(self):
        """记录现在的屏幕信息,以 numpy.array 格式覆盖保存到 self.screen"""
        self.screen = self.Android.snapshot()

    def get_screen(self, resolution=(1280, 720), need_screen_shot=True):
        if need_screen_shot:
            self.update_screen()
        return cv2.resize(self.screen, resolution)

    def get_pixel(self, x, y, screen_shot=False) -> list:
        """获取当前屏幕相对坐标 (x,y) 处的像素值
        Args:
            x (int): [0, 960)
            y (int): [0, 540)
        Returns:
            list[]: RGB 格式的像素值
        """
        if screen_shot:
            self.update_screen()
        if len(self.screen) != 540:
            self.screen = cv2.resize(self.screen, (960, 540))
        return [self.screen[y][x][2], self.screen[y][x][1], self.screen[y][x][0]]

    def check_pixel(self, position, bgr_color, distance=30, screen_shot=False) -> bool:
        """检查像素点是否满足要求
        Args:
            position (_type_): (x, y) 坐标, x 是长, 相对 960x540 的值, x \in [0, 960)

            bgr_color (_type_): 三元组, 顺序为 blue green red, 值为像素值

            distance (int, optional): 最大相差欧氏距离. Defaults to 30.

            screen_shot (bool, optional): 是否重新截图. Defaults to False.
        """
        color = self.get_pixel(*position, screen_shot)
        color.reverse()
        return CalcDis(color, bgr_color) < distance**2

    def locateCenterOnScreen(self, query: MyTemplate, confidence=0.85, this_methods=None):
        """从屏幕中找出和模板图像匹配度最高的矩阵区域的中心坐标
            参考 locateCenterOnImage
        Returns:
            如果找到返回一个二元组表示绝对坐标

            否则返回 None
        """
        if this_methods is None:
            this_methods = ["tpl"]
        return locateCenterOnImage(self.screen, query, confidence, this_methods)

    def get_image_position(self, image, need_screen_shot=1, confidence=0.85, this_methods=None):
        """从屏幕中找出和多张模板图像匹配度超过阈值的矩阵区域的中心坐标,如果有多个,返回第一个
            参考 locateCenterOnScreen
        Args:
            need_screen_shot (int, optional): 是否重新截取屏幕. Defaults to 1.
        Returns:
            如果找到:返回一个二元组表示相对坐标 (相对 960x540 屏幕)

            否则返回 None
        """
        if this_methods is None:
            this_methods = ["tpl"]
        images = image
        if not isinstance(images, Iterable):
            images = [images]
        if need_screen_shot == 1:
            self.update_screen()
        for image in images:
            res = self.locateCenterOnScreen(image, confidence, this_methods)
            if res is not None:
                return convert_position(res[0], res[1], self.config.resolution, mode="this_to_960")
        return None

    def get_images_position(self, images, need_screen_shot=1, confidence=0.85, this_methods=None):
        """同 get_image_position"""
        if this_methods is None:
            this_methods = ["tpl"]
        return self.get_image_position(images, need_screen_shot, confidence, this_methods)

    def image_exist(self, images, need_screen_shot=1, confidence=0.85, this_methods=None):
        """判断图像是否存在于屏幕中
        Returns:
            bool:如果存在为 True 否则为 False
        """
        if this_methods is None:
            this_methods = ["tpl"]
        if not isinstance(images, list):
            images = [images]
        if need_screen_shot:
            self.update_screen()
        return any(self.get_image_position(image, 0, confidence, this_methods) is not None for image in images)

    def images_exist(self, images, need_screen_shot=1, confidence=0.85, this_methods=None):
        """判断图像是否存在于屏幕中
        Returns:
            bool:如果存在为 True 否则为 False
        """
        if this_methods is None:
            this_methods = ["tpl"]
        return self.image_exist(images, need_screen_shot, confidence, this_methods)

    def wait_image(
        self,
        image: MyTemplate,
        confidence=0.85,
        timeout=10,
        gap=0.15,
        after_get_delay=0,
        this_methods=None,
    ):
        """等待一张图片出现在屏幕中,置信度超过一定阈值(支持多图片)

        Args:
            timeout (int, optional): 最大等待时间. Defaults to 10.
        Returns:
            如果在 timeout 秒内发现,返回一个二元组表示其相对(960x540 屏幕)位置

            否则返回 False
        """
        if this_methods is None:
            this_methods = ["tpl"]
        if timeout < 0:
            raise ValueError("arg 'timeout' should at least be 0 but is ", str(timeout))
        StartTime = time.time()
        while True:
            x = self.get_image_position(image, 1, confidence, this_methods)
            if x != None:
                time.sleep(after_get_delay)
                return x
            if time.time() - StartTime > timeout:
                time.sleep(gap)
                return False
            time.sleep(gap)

    def wait_images(self, images=None, confidence=0.85, gap=0.15, after_get_delay=0, timeout=10):
        """等待一系列图片中的一个在屏幕中出现

        Args:
            images (list, optional): 很多图片,可以是列表或字典. Defaults to [].
            confidence (_type_, optional): 置信度. Defaults to 0.85.
            timeout (int, optional): 最长等待时间. Defaults to 10.

        Raises:
            TypeError: image_list 中有不合法参数

        Returns:
            None: 未找到任何图片
            a number of int: 第一个出现的图片的下标(0-based) if images is a list
            the key of the value: if images is a dict
        """
        if timeout < 0:
            raise ValueError("arg 'timeout' should at least be 0 but is ", str(timeout))
        if images is None:
            return None

        if isinstance(images, MyTemplate):
            images = [(0, images)]
        elif isinstance(images, (list, Tuple)) and isinstance(images[0], MyTemplate):
            images = list(enumerate(images))
        elif isinstance(images, dict):
            images = images.items()

        StartTime = time.time()
        while True:
            self.update_screen()
            for res, image in images:
                if self.image_exist(image, False, confidence):
                    time.sleep(after_get_delay)
                    return res
            time.sleep(gap)
            if time.time() - StartTime > timeout:
                return None

    def wait_images_position(self, images=None, confidence=0.85, gap=0.15, after_get_delay=0, timeout=10):
        """等待一些图片,并返回第一个匹配结果的位置

        参考 wait_images
        """
        if images is None:
            images = []
        if not isinstance(images, Iterable):
            images = [images]
        rank = self.wait_images(images, confidence, gap, after_get_delay, timeout)
        if rank is None:
            return None
        return self.get_image_position(images[rank], 0, confidence)

    def click_image(self, image, must_click=False, timeout=0, delay=0.5):
        """点击一张图片的中心位置

        Args:
            image (MyTemplate): 目标图片

            must_click (bool, optional): 如果为 True,点击失败则抛出异常. Defaults to False.

            timeout (int, optional): 等待延时. Defaults to 0.

            delay (float, optional): 点击后延时. Defaults to 0.5.

        Raises:
            NotFoundErr: 如果在 timeout 时间内未找到则抛出该异常
        """
        pos = self.wait_images_position(image, gap=0.03, timeout=timeout)
        if pos == False:
            if must_click == False:
                return False
            else:
                raise ImageNotFoundErr(f"Target image not found:{str(image.filepath)}")

        self.Android.click(*pos, delay=delay)
        return True

    def click_images(self, images, must_click=False, timeout=0, delay=0.5):
        """点击一些图片中第一张出现的,如果有多个,点击第一个"""
        self.click_image(images, must_click, timeout)

    def log_screen(self, need_screen_shot=False):
        """向默认数据记录路径记录当前屏幕数据,带时间戳保存,960x540大小
        Args:
            need_screen_shot (bool, optional): 是否新截取一张图片. Defaults to False.
        """
        if need_screen_shot:
            self.update_screen()
        screen = copy.deepcopy(self.screen)
        screen = cv2.resize(screen, (960, 540))
        self.logger.log_image(image=screen, name=datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))



import datetime
import os
from types import SimpleNamespace as SN

import keyboard as kd

import autowsgr
from autowsgr.controller.run_timer import Emulator, Timer
from autowsgr.utils.io import recursive_dict_update, yaml_to_dict
from autowsgr.utils.new_logger import Logger

event_pressed = set()
script_end = 0


def start_script(settings_path=None):
    """启动脚本, 返回一个 Timer 记录器.
    :如果模拟器没有运行, 会尝试启动模拟器,
    :如果游戏没有运行, 会自动启动游戏,
    :如果游戏在后台, 会将游戏转到前台
    Returns:
        Timer: 该模拟器的记录器
    """
    config = yaml_to_dict(os.path.join(os.path.dirname(autowsgr.__file__), "data", "default_settings.yaml"))
    if settings_path is not None:
        user_settings = yaml_to_dict(settings_path)
        config = recursive_dict_update(config, user_settings)
    else:
        print("========Warning========")
        print(
            f"No user_settings file specified, default settings {os.path.join(os.path.dirname(autowsgr.__file__), 'data', 'default_settings.yaml')} will be used."
        )
        print("=========End===========")

    # set logger
    config["log_dir"] = os.path.join(config["LOG_PATH"], datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
    os.makedirs(config["log_dir"], exist_ok=True)
    logger = Logger(config)
    config = SN(**config)
    timer = Timer(config, logger)
    config_str = logger.save_config(config)
    timer.logger.reset_level()  # sb airtest reset all the fucking log level
    return timer


def start_script_emulator(settings_path=None):
    """启动脚本, 返回一个 Emulator 记录器, 用于操作模拟器
    Returns:
        Emulator: 该模拟器的记录器
    """
    config = yaml_to_dict(os.path.join(os.path.dirname(autowsgr.__file__), "data", "default_settings.yaml"))
    if settings_path is not None:
        user_settings = yaml_to_dict(settings_path)
        config = recursive_dict_update(config, user_settings)
    else:
        print("========Warning========")
        print(
            f"No user_settings file specified, default settings {os.path.join(os.path.dirname(autowsgr.__file__), 'data', 'default_settings.yaml')} will be used."
        )
        print("=========End===========")

    # set logger
    config["log_dir"] = os.path.join(config["LOG_PATH"], datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
    os.makedirs(config["log_dir"], exist_ok=True)
    logger = Logger(config)
    config = SN(**config)
    enmulator = Emulator(config, logger)
    config_str = logger.save_config(config)
    enmulator.logger.reset_level()  # sb airtest reset all the fucking log level
    return enmulator



import functools
import os
import subprocess

import cv2
import easyocr

from autowsgr.constants.data_roots import TUNNEL_ROOT
from autowsgr.utils.operator import unzip_element

en_reader = None
ch_reader = None


def load_en_reader():
    global en_reader
    en_reader = easyocr.Reader(["en"], gpu=True)


def load_ch_reader():
    global ch_reader
    ch_reader = easyocr.Reader(["ch_sim"], gpu=True)


def edit_distance(word1, word2) -> int:
    """
    解题思路，动态规划
        步骤1：将word1与word2前拼接上空格，方便为空时的操作
        步骤2：初始化dp第一个元素，接着初始化dp的第一行与第一列
        步骤3：可通过画表(如: excel里)找到状态转移的规律，填充剩下的dp格子即可
    :return: dp[-1][-1], 返回最后操作的结果
    """
    m, n = len(word1), len(word2)
    if m == 0 and n == 0:
        return 0
    word1, word2 = " " + word1, " " + word2  # 非常必要的操作，不添加空格话，在Word为空时会比较麻烦
    dp = [[0 for _ in range(n + 1)] for _ in range(m + 1)]
    dp[0][0] = 0  # 初始化dp[0][0] = 0，因为空格对空格不需要任何操作，即0步
    for i in range(1, n + 1):  # 第一行初始化
        dp[0][i] = i
    for i in range(1, m + 1):  # 第一列初始化
        dp[i][0] = i
    for i in range(1, m + 1):  # 逐个填充剩余的dp格子
        for j in range(1, n + 1):
            if word1[i] == word2[j]:
                dp[i][j] = dp[i - 1][j - 1]
            else:
                dp[i][j] = min(dp[i - 1][j - 1], dp[i - 1][j], dp[i][j - 1]) + 1
    return dp[-1][-1]


def find_lcseque(s1, s2):
    """求两个字符串的LCS"""
    m = [[0 for x in range(len(s2) + 1)] for y in range(len(s1) + 1)]
    d = [[None for x in range(len(s2) + 1)] for y in range(len(s1) + 1)]
    for p1 in range(len(s1)):
        for p2 in range(len(s2)):
            if s1[p1] == s2[p2]:
                m[p1 + 1][p2 + 1] = m[p1][p2] + 1
                d[p1 + 1][p2 + 1] = "ok"
            elif m[p1 + 1][p2] > m[p1][p2 + 1]:
                m[p1 + 1][p2 + 1] = m[p1 + 1][p2]
                d[p1 + 1][p2 + 1] = "left"
            else:
                m[p1 + 1][p2 + 1] = m[p1][p2 + 1]
                d[p1 + 1][p2 + 1] = "up"

    (p1, p2) = (len(s1), len(s2))
    s = []
    while m[p1][p2]:  # 不为None时
        c = d[p1][p2]
        if c == "ok":  # 匹配成功，插入该字符，并向左上角找下一个
            s.append(s1[p1 - 1])
            p1 -= 1
            p2 -= 1
        if c == "left":  # 根据标记，向左找下一个
            p2 -= 1
        if c == "up":  # 根据标记，向上找下一个
            p1 -= 1
    s.reverse()
    return "".join(s)


def get_allow(names):
    char_set = set()
    for name in unzip_element(names):
        for char in name:
            char_set.add(char)
    res = ""
    for char in char_set:
        res += char
    return res


def replace(origin_str) -> str:
    """一些不可避免的识别错误的替换规则
    Args:
        origin_str (str): 原始字符串
    """
    if origin_str[0] == "0" and len(origin_str) >= 3:  # U 艇识别为 0xxx
        origin_str = "U" + origin_str[1:]
    return origin_str


def compare_box(A, B):
    """对 easyocr.readtext 返回的 box 进行排序"""
    k1, k2 = A[0][0], B[0][0]
    if abs(k1[0] - k2[0]) <= 5 and abs(k1[1] - k2[1]) <= 5:
        return 0
    if abs(k1[1] - k2[1]) > 5:
        if k1[1] < k2[1]:
            return -1
        else:
            return 1
    else:
        if k1[0] < k2[1]:
            return -1
        else:
            return 1


def recognize(image, char_list, min_size=7, text_threshold=0.55, low_text=0.3):
    if ch_reader == None:
        load_ch_reader()
    result = ch_reader.readtext(
        image,
        allowlist=char_list,
        min_size=min_size,
        text_threshold=text_threshold,
        low_text=low_text,
    )
    # print(result)
    return result


def recognize_single_number(image, ex_list="", min_size=7, text_threshold=0.55, low_text=0.3):
    """识别单个数字
    Returns:
        int: 结果
        False: 识别失败
    """
    result = recognize_number(image, ex_list, min_size, text_threshold, low_text)
    if len(result) != 1:
        return False
    return int(result[0][1])


def recognize_number(image, ex_list="", min_size=7, text_threshold=0.55, low_text=0.3):
    """识别数字和 ex_list 的字符, 返回识别结果列表,
    Returns:
        list(result): 一个 result 为 [position, text, confidence]
    """
    if en_reader == None:
        load_en_reader()
    char_list = "0123456789"
    for ch in ex_list:
        if char_list.find(ch) == -1:
            char_list += ch
    result = en_reader.readtext(
        image,
        allowlist=char_list,
        min_size=min_size,
        text_threshold=text_threshold,
        low_text=low_text,
    )
    # print(result)
    return result


def _recognize_ship(image, names, char_list=None, min_size=7, text_threshold=0.55, low_text=0.3):
    """识别没有预处理过的图片中的舰船, 返回识别结果列表,
    Returns:
        list(result): 一个 result 为 [ship_name, left_top]
    """
    if char_list is None:
        char_list = get_allow(names)
    if ch_reader == None:
        load_ch_reader()
    result = recognize(
        image,
        char_list=char_list,
        min_size=min_size,
        text_threshold=text_threshold,
        low_text=low_text,
    )
    result = [x for x in result if x[1] != ""]  # 去除空匹配
    results = []
    sorted(result, key=functools.cmp_to_key(compare_box))
    for box in result:
        res, lcs, name = "dsjfagiahsdifhaoisd", "", box[1]
        name = replace(name)
        for _name in names:
            if any([(_name.find(char) != -1) for char in name]):
                dis1 = edit_distance(_name, name)
                dis2 = edit_distance(res, name)
                if dis1 < dis2:
                    res = _name
        results.append((res, box[0]))
    if len(results) == 0:
        results.append(("Unknown", (0, 0)))
    # print(results)
    return results


def recognize_ship(image, names, char_list=None, min_size=7, text_threshold=0.55, low_text=0.3):
    """传入一张图片,返回舰船信息,包括名字和舰船型号

    Args:
        image (_type_): _description_
        names (_type_): _description_
        char_list (_type_, optional): _description_. Defaults to None.
    """
    if isinstance(image, str):
        image_path = os.path.abspath(image)
    else:
        image_path = os.path.join(TUNNEL_ROOT, "OCR.PNG")
        cv2.imwrite(image_path, image)
    if char_list is None:
        char_list = get_allow(names)
    with open(os.path.join(TUNNEL_ROOT, "locator.in"), "w+") as f:
        f.write(image_path)
    locator_exe = os.path.join(TUNNEL_ROOT, "locator.exe")
    subprocess.run([locator_exe, TUNNEL_ROOT])
    if os.path.exists(os.path.join(TUNNEL_ROOT, "1.PNG")):
        result = _recognize_ship(
            os.path.join(TUNNEL_ROOT, "1.PNG"),
            names,
            char_list,
            min_size=min_size,
            text_threshold=text_threshold,
            low_text=low_text,
        )
    else:
        result = _recognize_ship(
            "1.PNG",
            names,
            char_list,
            min_size=min_size,
            text_threshold=text_threshold,
            low_text=low_text,
        )
    # print(result)
    return result


def recover(image):
    for p in image:
        pass


if __name__ == "__main__":
    pass



import os

from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.controller.run_timer import Timer
from autowsgr.fight.battle import BattleInfo, BattlePlan
from autowsgr.fight.common import start_march
from autowsgr.fight.event.event import PatrollingEvent
from autowsgr.fight.normal_fight import NormalFightInfo, NormalFightPlan
from autowsgr.game.game_operation import MoveTeam, detect_ship_stats, quick_repair
from autowsgr.utils.math_functions import CalcDis

MAP_POSITIONS = [
    None,
    (275, 118),
    (364, 383),
    (578, 147),
    (522, 337),
    (445, 158),
    (791, 157),
]


class EventFightPlan20220928(PatrollingEvent, NormalFightPlan):
    def __init__(self, timer: Timer, plan_path, fleet_id=1, event="20220928"):
        self.event_name = event
        NormalFightPlan.__init__(self, timer, plan_path, fleet_id=fleet_id)
        PatrollingEvent.__init__(self, timer, event, MAP_POSITIONS)

    def _load_fight_info(self):
        self.Info = EventFightInfo20220928(self.timer, self.chapter, self.map)
        self.Info.load_point_positions(os.path.join(MAP_ROOT, "event", self.event_name))

    def _change_fight_map(self, chapter_id, map_id):
        self.enter_map(chapter_id, map_id)
        while self.timer.image_exist(self.common_image["little_monster"]) == False:  # 找到小怪物图标,点击下方进入主力决战
            while self.timer.wait_images_position(self.common_image["monster"], timeout=1) is None:
                self.random_walk()
            if self.timer.image_exist(self.common_image["little_monster"]):
                break
            ret = self.timer.wait_images_position(self.common_image["monster"])
            self.timer.Android.click(*ret)

        ret = self.get_close(self.common_image["little_monster"][0])
        self.timer.Android.click(ret[0], ret[1] + 60)
        assert self.timer.wait_image(self.event_image[1])


class EventFightInfo20220928(PatrollingEvent, NormalFightInfo):
    def __init__(self, timer: Timer, chapter_id, map_id, event="20220928") -> None:
        NormalFightInfo.__init__(self, timer, chapter_id, map_id)
        PatrollingEvent.__init__(self, timer, event, MAP_POSITIONS)
        self.map_image = self.event_image[1]
        self.end_page = "unknown_page"
        self.state2image["map_page"] = [self.map_image, 5]


class EventFightPlan20220928_2(PatrollingEvent, BattlePlan):
    def __init__(self, timer: Timer, plan_path, fleet_id=1, event="20220928"):
        self.event_name = event
        self.fleet_id = fleet_id
        BattlePlan.__init__(self, timer, plan_path)
        PatrollingEvent.__init__(self, timer, event, MAP_POSITIONS)

    def _enter_fight(self, same_work=False):
        self._go_map_page()
        self.enter_map(self.chapter, self.map)
        assert self.timer.wait_image(self.event_image[2]) is not False
        self.find(self.enemy_image[self.target])
        ret = self.get_close(self.enemy_image[self.target])
        self.timer.Android.click(*ret)
        assert self.timer.wait_image(self.event_image[4]) != False
        self.timer.Android.click(900, 500)
        self.timer.wait_pages("fight_prepare_page", after_wait=0.15)
        MoveTeam(self.timer, self.fleet_id)
        self.Info.ship_stats = detect_ship_stats(self.timer)
        quick_repair(self.timer, self.repair_mode, self.Info.ship_stats)
        return start_march(self.timer)


class EventFightInfo20220928_2(PatrollingEvent, BattleInfo):
    def __init__(self, timer: Timer, chapter, map, event="20220928") -> None:
        BattleInfo.__init__(self, timer, chapter, map)
        PatrollingEvent.__init__(self, timer, event, MAP_POSITIONS)
        self.end_page = "unknown_page"
        self.state2image["battle_page"] = [self.event_image[2], 5]



import json
import logging
import os
import sys

from autowsgr.utils.io import save_image

# import streamlit as st


class Logger:
    def __init__(self, config):
        self.config = config
        self.log_dir = config["log_dir"]
        if "log_level" in config.keys():
            log_level = config["log_level"]
        else:
            log_level = "INFO"
        self.log_level = log_level
        self.console_logger = self._get_logger(log_level)

    def save_config(self, config):
        # write config file
        config_str = json.dumps(vars(config), ensure_ascii=False, indent=4, sort_keys=True)
        with open(os.path.join(self.log_dir, "config.json"), "w") as f:
            f.write(config_str)
        return config_str

    def reset_level(self):
        self.console_logger.setLevel(self.log_level)

    def debug(self, *args):
        self.console_logger.debug(str(args))

    def info(self, string):
        self.console_logger.info(string)
        # st.write(string)

    def warning(self, string):
        self.console_logger.warning("===================WARNING===================")
        self.console_logger.warning(string)
        self.console_logger.warning("====================END====================")
        # st.write(string)

    def error(self, string):
        self.console_logger.error("===================ERROR===================")
        self.console_logger.error(string)
        self.console_logger.error("====================END====================")
        # st.write(string)

    def log_stat(self, key, value, t, tag="train"):
        self.info(f"{tag} {key}: {value:.4f}")

    def log_image(
        self,
        image,
        name,
        ndarray_mode="BGR",
        ignore_existed_image=False,
        *args,
        **kwargs,
    ):
        """向默认数据记录路径记录图片
        Args:
            image: 图片,PIL.Image.Image 格式或者 numpy.ndarray 格式
            name (str): 图片文件名
        """
        if "png" not in name and "PNG" not in name:
            name += ".PNG"
        path = os.path.join(self.log_dir, name)

        save_image(
            path=path,
            image=image,
            ignore_existed_image=ignore_existed_image,
            *args,
            **kwargs,
        )

    def _get_logger(self, log_level="INFO") -> logging.Logger:
        logger = logging.getLogger("autowsgr")
        logger.propagate = False
        logger.handlers = []
        logger.setLevel(log_level)

        # Stream
        ch_formatter = logging.Formatter("[%(levelname)s %(asctime)s %(name)s] %(message)s", "%H:%M:%S")
        ch = logging.StreamHandler(sys.stderr)
        ch.setFormatter(ch_formatter)
        logger.addHandler(ch)

        # File
        self.log_file_path = os.path.join(self.log_dir, "console.log")
        ch = logging.FileHandler(self.log_file_path, encoding="utf-8")
        ch.setFormatter(ch_formatter)
        logger.addHandler(ch)

        return logger



import math
import os
import subprocess
import time

import numpy as np
from PIL import Image as PIM

from autowsgr.constants.colors import COLORS
from autowsgr.constants.data_roots import TUNNEL_ROOT
from autowsgr.constants.image_templates import IMG
from autowsgr.constants.other_constants import (
    AADG,
    ASDG,
    AV,
    BB,
    BBV,
    BC,
    BG,
    BM,
    CA,
    CAV,
    CBG,
    CL,
    CLT,
    CV,
    CVL,
    DD,
    NAP,
    NO,
    RESOURCE_NAME,
    SAP,
    SC,
    SS,
)
from autowsgr.constants.positions import BLOOD_BAR_POSITION, TYPE_SCAN_AREA
from autowsgr.controller.run_timer import Timer
from autowsgr.ocr.digit import get_resources
from autowsgr.utils.io import delete_file, read_file
from autowsgr.utils.math_functions import CalcDis, CheckColor, matrix_to_str


class Resources:
    def __init__(self, timer: Timer):
        self.timer = timer
        self.resources = {}

    def detect_resources(self, name=None):
        timer = self.timer
        if name is not None:
            if name in ("normal", "oil", "ammo", "steel", "aluminum"):
                self.timer.goto_game_page("main_page")
                self.detect_resources()
            if name == "quick_repair":
                self.timer.goto_game_page("choose_repair_page")
                self.detect_resources()
            if name == "quick_build":
                self.timer.goto_game_page("build_page")
                self.detect_resources()
            if name == "ship_blueprint":
                self.timer.goto_game_page("build_page")
                self.detect_resources()
            if name == "equipment_blueprint":
                self.timer.goto_game_page("develop_page")
                self.detect_resources()
        else:
            result = get_resources(timer)
            for key, value in result.items():
                self.resources[key] = value

    def ask_resources(self, name, detect=False):
        """查询资源量(不会从游戏中探查,会根据程序维护的数据返回)
        如果游戏脱离程序监控,可能会不准确，需要先 detect

        Args:
            name (资源名称):
                values:
                    refer to constants.other_constants.RESOURCE_NAME
            detect (bool, optional): 是否从游戏中重新探查(如果否，则使用由程序维护的数据)

        Returns:
            int: 资源量
        """
        if name not in RESOURCE_NAME:
            raise ValueError("Unsupported resource name")
        if detect or name not in self.resources.keys():
            self.detect_resources(name)

        return self.resources.get(name)


def get_enemy_condition(timer: Timer, type="exercise", *args, **kwargs):
    """获取敌方舰船类型数据并返回一个字典, 具体图像识别为黑箱, 采用 C++ 实现

    Args:
        type (str, optional): 描述情景. Defaults to 'exercise'.
            'exercise': 演习点击挑战后的一横排舰船

            'fight': 索敌成功后的两列三行

    Returs:
        dict: {[SHIP_TYPE]:[SHIP_AMOUNT]}

        example return: {"CV":1, "BB":3, "DD": 2}
    """

    enemy_type_count = {
        CV: 0,
        BB: 0,
        SS: 0,
        BC: 0,
        NAP: 0,
        DD: 0,
        ASDG: 0,
        AADG: 0,
        CL: 0,
        CA: 0,
        CLT: 0,
        CVL: 0,
        NO: 0,
        "all": 0,
        CAV: 0,
        AV: 0,
        BM: 0,
        SAP: 0,
        BG: 0,
        CBG: 0,
        SC: 0,
        BBV: 0,
        "AP": 0,
    }

    if type == "exercise":
        type = 0
    if type == "fight":
        type = 1

    if timer.image_exist(IMG.fight_image[12]):
        # 特殊补给舰
        enemy_type_count[SAP] = 1

    # 处理图像并将参数传递给识别图像的程序
    img = PIM.fromarray(timer.screen).convert("L")
    img = img.resize((960, 540))
    input_path = os.path.join(TUNNEL_ROOT, "args.in")
    output_path = os.path.join(TUNNEL_ROOT, "res.out")
    delete_file(output_path)
    args = "recognize\n6\n"
    for i, area in enumerate(TYPE_SCAN_AREA[type]):
        arr = np.array(img.crop(area))
        args += matrix_to_str(arr)
    with open(input_path, "w") as f:
        f.write(args)
    recognize_enemy_exe = os.path.join(TUNNEL_ROOT, "recognize_enemy.exe")
    subprocess.run([recognize_enemy_exe, TUNNEL_ROOT])

    # 获取并解析结果
    res = read_file(os.path.join(TUNNEL_ROOT, "res.out")).split()
    enemy_type_count["ALL"] = 0
    for i, x in enumerate(res):
        enemy_type_count[x] += 1
        if x != NO:
            enemy_type_count["ALL"] += 1
    enemy_type_count[NAP] = enemy_type_count["AP"] - enemy_type_count[SAP]
    count = {}
    for key, value in enemy_type_count.items():
        if value:
            count[key] = value
    timer.logger.debug("enemys:" + str(count))
    return count


def detect_ship_stats(timer: Timer, type="prepare", previous=None):
    """检查我方舰船的血量状况(精确到红血黄血绿血)并返回

    Args:
        type (str, optional): 描述在哪个界面检查. .
            'prepare': 战斗准备界面

            'sumup': 单场战斗结算界面
    Returns:
        list: 表示血量状态

        example: [-1, 0, 0, 1, 1, 2, -1] 表示 1-2 号位绿血 3-4 号位中破, 5 号位大破, 6 号位不存在

    """
    # Todo: 检测是否满血/触发中保, 精确到数值的检测, 战斗结算时检测不依赖先前信息

    timer.update_screen()
    result = [-1, 0, 0, 0, 0, 0, 0]
    for i in range(1, 7):
        if type == "prepare":
            pixel = timer.get_pixel(*BLOOD_BAR_POSITION[0][i])
            result[i] = CheckColor(pixel, COLORS.BLOOD_COLORS[0])
            if result[i] in [3, 2]:
                result[i] = 2
            elif result[i] == 0:
                result[i] = 0
            elif result[i] == 4:
                result[i] = -1
            else:
                result[i] = 1
        elif type == "sumup":
            if previous and previous[i] == -1:
                result[i] = -1
                continue
            pixel = timer.get_pixel(*BLOOD_BAR_POSITION[1][i])
            result[i] = CheckColor(pixel, COLORS.BLOOD_COLORS[1])
            if result[i] == 4:
                result[i] = -1
    return result


def detect_ship_type(timer: Timer):
    """ToDo
    在出征准备界面读取我方所有舰船类型并返回该列表
    """


def get_exercise_stats(timer: Timer, robot=None):
    """检查演习界面, 第 position 个位置,是否为可挑战状态, 强制要求屏幕中只有四个目标

    Args:
        position (_type_): 编号从屏幕中的位置编起, 如果滑动显示了第 2-5 个敌人, 那么第二个敌人编号为 1

    Returns:
        bool: 如果可挑战, 返回 True , 否则为 False, 1-index
    """
    timer.update_screen()
    up = timer.check_pixel((933, 59), (177, 171, 176), distance=60)
    down = timer.check_pixel((933, 489), (177, 171, 176), distance=60)
    assert (up and down) == False

    result = [
        None,
    ]
    if up == False and down == False:
        timer.Android.swipe(800, 200, 800, 400)  # 上滑
        timer.update_screen()
        up = True
    if up:
        for position in range(1, 5):
            result.append(math.sqrt(CalcDis(timer.get_pixel(770, position * 110 - 10), COLORS.CHALLENGE_BLUE)) <= 50)
        timer.Android.swipe(800, 400, 800, 200)  # 下滑
        timer.update_screen()
        result.append(math.sqrt(CalcDis(timer.get_pixel(770, 4 * 110 - 10), COLORS.CHALLENGE_BLUE)) <= 50)
        return result
    if down:
        for position in range(1, 5):
            result.append(math.sqrt(CalcDis(timer.get_pixel(770, position * 110 - 10), COLORS.CHALLENGE_BLUE)) <= 50)
        if robot is not None:
            result.insert(1, robot)
        else:
            timer.Android.swipe(800, 200, 800, 400)  # 上滑
            timer.update_screen()
            result.insert(
                1,
                math.sqrt(CalcDis(timer.get_pixel(770, 4 * 110 - 10), COLORS.CHALLENGE_BLUE)) <= 50,
            )

            timer.Android.swipe(800, 400, 800, 200)  # 下滑

        return result


def check_support_stats(timer: Timer):
    """在出征准备界面检查是否开启了战役支援(有开始出征按钮的界面)

    Returns:
        bool: 先判断是否为灰色，如果为灰色则返回True，然后判断是否开启，如果开启则返回True，否则返回False
    """
    timer.update_screen()
    pixel = timer.get_pixel(623, 75)
    d1 = CalcDis(pixel, COLORS.SUPPORT_ENABLE)  # 支援启用的黄色
    d2 = CalcDis(pixel, COLORS.SUPPORT_DISABLE)  # 支援禁用的蓝色
    d3 = CalcDis(pixel, COLORS.SUPPORT_ENLESS)  # 支援次数用尽的灰色
    if d1 > d3 and d2 > d3:
        timer.logger.info("战役支援次数已用尽")
        return True
    else:
        return d1 < d2



import os

from autowsgr.constants import literals
from autowsgr.constants.image_templates import IMG
from autowsgr.controller.run_timer import Timer
from autowsgr.game.game_operation import get_ship, quick_repair
from autowsgr.game.get_game_info import detect_ship_stats
from autowsgr.utils.io import recursive_dict_update, yaml_to_dict

from .common import DecisionBlock, FightInfo, FightPlan, start_march

"""
战役模块/单点战斗模板
"""


class BattleInfo(FightInfo):
    def __init__(self, timer: Timer) -> None:
        super().__init__(timer)

        self.end_page = "battle_page"

        self.successor_states = {
            "proceed": ["spot_enemy_success", "formation", "fight_period"],
            "spot_enemy_success": {
                "retreat": ["battle_page"],
                "fight": ["formation", "fight_period"],
            },
            "formation": ["fight_period"],
            "fight_period": ["night", "result"],
            "night": {
                "yes": ["result"],
                "no": [["result", 7]],
            },
            "night_fight_period": ["result"],
            "result": ["battle_page"],  # 两页战果
        }

        self.state2image = {
            "proceed": [IMG.fight_image[5], 7.5],
            "spot_enemy_success": [IMG.fight_image[2], 15],
            "formation": [IMG.fight_image[1], 15, 0.8],
            "fight_period": [IMG.symbol_image[4], 7.5],
            "night": [IMG.fight_image[6], 150],
            "result": [IMG.fight_image[16], 75],
            "battle_page": [
                IMG.identify_images["battle_page"][0],
                7.5,
            ],
        }

        self.after_match_delay = {
            "night": 1.75,
            "get_ship": 1,
        }

    def reset(self):
        self.fight_history.reset()
        self.last_state = ""
        self.last_action = ""
        self.state = "proceed"

    def _before_match(self):
        # 点击加速
        if self.state in ["proceed"]:
            self.timer.Android.click(380, 520, delay=0, enable_subprocess=True, not_show=True)
        self.timer.update_screen()

    def _after_match(self):
        if self.state == "get_ship":
            get_ship(self.timer)
        super()._after_match()


class BattlePlan(FightPlan):
    def __init__(self, timer, plan_path=None) -> None:
        super().__init__(timer)

        # 加载默认配置
        default_args = yaml_to_dict(os.path.join(self.config.PLAN_ROOT, "default.yaml"))
        plan_defaults = default_args["battle_defaults"]
        plan_defaults.update({"node_defaults": default_args["node_defaults"]})

        # 加载计划配置
        if plan_path is not None:
            plan_args = yaml_to_dict(os.path.join(self.config.PLAN_ROOT, plan_path))
            args = recursive_dict_update(plan_defaults, plan_args, skip=["node_args"])
        else:
            args = plan_defaults
        self.__dict__.update(args)

        # 加载节点配置
        node_defaults = self.node_defaults
        if plan_path is not None:
            node_args = recursive_dict_update(node_defaults, plan_args["node_args"])
        else:
            node_args = node_defaults
        self.node = DecisionBlock(timer, node_args)
        self.Info = BattleInfo(timer)

    def _go_fight_prepare_page(self):
        self.timer.goto_game_page("battle_page")
        now_hard = self.timer.wait_images([IMG.fight_image[9], IMG.fight_image[15]])
        hard = self.map > 5
        if now_hard != hard:
            self.timer.Android.click(800, 80, delay=1)

    def _enter_fight(self, same_work=False) -> str:
        if same_work == False:
            self._go_fight_prepare_page()
        self.timer.Android.click(180 * ((self.map - 1) % 5 + 1), 200)
        self.timer.wait_pages("fight_prepare_page", after_wait=0.15)
        self.Info.ship_stats = detect_ship_stats(self.timer)
        quick_repair(self.timer, self.repair_mode, ship_stats=self.Info.ship_stats)
        return start_march(self.timer)

    def _make_decision(self) -> str:
        self.update_state()
        if self.Info.state == "battle_page":
            return literals.FIGHT_END_FLAG

        # 进行通用 NodeLevel 决策
        action, fight_stage = self.node.make_decision(self.Info.state, self.Info.last_state, self.Info.last_action, self.Info)
        self.Info.last_action = action
        return fight_stage



import os

from autowsgr.constants.data_roots import MAP_ROOT
from autowsgr.controller.run_timer import Timer
from autowsgr.fight.common import start_march
from autowsgr.fight.event.event import Event
from autowsgr.fight.normal_fight import NormalFightInfo, NormalFightPlan
from autowsgr.game.game_operation import MoveTeam, quick_repair
from autowsgr.utils.math_functions import CalcDis

NODE_POSITION = (
    None,
    (168, 354),
    (83, 221),
    (280, 161),
    (670, 139),
    (766, 374),
    (540, 300),
)


class EventFightPlan20221118(Event, NormalFightPlan):
    def __init__(self, timer: Timer, plan_path, fleet_id=None, event="20221118"):
        """
        Args:
            fleet_id : 新的舰队参数, 如果为 None 则使用计划参数.
        """
        self.event_name = event
        NormalFightPlan.__init__(self, timer, plan_path, fleet_id=fleet_id)
        Event.__init__(self, timer, event)

    def _load_fight_info(self):
        self.Info = EventFightInfo20221118(self.timer, self.chapter, self.map)
        self.Info.load_point_positions(os.path.join(MAP_ROOT, "event", self.event_name))

    def _change_fight_map(self, chapter_id, map_id):
        """选择并进入战斗地图(chapter-map)"""
        self.change_difficulty(chapter_id)

    def _go_fight_prepare_page(self) -> None:
        if not self.timer.image_exist(self.Info.event_image[1], need_screen_shot=0):
            self.timer.Android.click(*NODE_POSITION[self.map])
        self.timer.wait_image(self.event_image[1])
        self.timer.Android.click(850, 490)
        self.timer.wait_pages("fight_prepare_page", after_wait=0.15)


class EventFightInfo20221118(Event, NormalFightInfo):
    def __init__(self, timer: Timer, chapter_id, map_id, event="20221118") -> None:
        NormalFightInfo.__init__(self, timer, chapter_id, map_id)
        Event.__init__(self, timer, event)
        self.map_image = self.common_image["easy"] + self.common_image["hard"] + [self.event_image[1]]
        self.end_page = "unknown_page"
        self.state2image["map_page"] = [self.map_image, 5]



import os
import re

from airtest.core.cv import (
    MATCHING_METHODS,
    ST,
    InvalidMatchingMethodError,
    TargetPos,
    Template,
)
from airtest.core.settings import Settings as ST

from .data_roots import IMG_ROOT


class MyTemplate(Template):
    def match_in(self, screen, this_methods=None):
        match_result = self._cv_match(screen, this_methods)
        if not match_result:
            return None
        focus_pos = TargetPos().getXY(match_result, self.target_pos)
        return focus_pos

    def _cv_match(self, screen, this_methods=None):
        ori_image = self._imread()
        image = self._resize_image(ori_image, screen, ST.RESIZE_METHOD)
        ret = None
        if this_methods is None:
            this_methods = ST.CVSTRATEGY
        for method in this_methods:
            # get function definition and execute:
            func = MATCHING_METHODS.get(method, None)
            if func is None:
                raise InvalidMatchingMethodError(
                    f"Undefined method in CVSTRATEGY: '{method}', try 'kaze'/'brisk'/'akaze'/'orb'/'surf'/'sift'/'brief' instead."
                )
            if method in ["mstpl", "gmstpl"]:
                ret = self._try_match(
                    func,
                    ori_image,
                    screen,
                    threshold=self.threshold,
                    rgb=self.rgb,
                    record_pos=self.record_pos,
                    resolution=self.resolution,
                    scale_max=self.scale_max,
                    scale_step=self.scale_step,
                )
            else:
                ret = self._try_match(func, image, screen, threshold=self.threshold, rgb=self.rgb)
            if ret:
                break
        return ret


def make_dir_templates(path):
    """建立 path 目录下所有图片的模板字典"""
    # 不处理二级目录和非 .png 文件
    all_files = [file for file in os.listdir(path) if not (os.path.isdir(file or file.split(".")[-1].lower() != "png"))]

    if all(file.split(".")[0].isdecimal() for file in all_files):
        res = [
            None,
        ]
        for file in all_files:
            key = int(file.split(".")[0])
            if key >= len(res):
                res.extend(
                    [
                        None,
                    ]
                    * (key - len(res) + 1)
                )
            file_path = os.path.join(path, file)
            res[key] = MyTemplate(file_path, 0.9, resolution=(960, 540))
    else:
        res = {}
        for file in all_files:
            file_path = os.path.join(path, file)
            key_name = file.split(".")[0]
            if key_name.isdecimal():
                res[int(key_name)] = MyTemplate(file_path, 0.9, resolution=(960, 540))
            res[key_name] = MyTemplate(file_path, 0.9, resolution=(960, 540))

    return res


def make_dir_templates_without_number(path):
    """给定路径, 返回一个字典

    字典的键为路径下所有图片的英文字母文件名(不含后缀)

    字典的值为对应图片的 MyTemplate 类列表, 所有英文名相同的图片会被放入同一个列表(仅忽略数字)
    """
    res = {}
    for file in os.listdir(path):
        filename = file.split(".")[0]
        key_name = re.findall(r"[^0-9]*", filename)[0]
        if key_name not in res.keys():
            res[key_name] = []
        file_path = os.path.join(path, file)
        res[key_name].append(MyTemplate(file_path, 0.9, resolution=(960, 540)))
    return res


class ImageSet:
    def __init__(self):
        # identify_images 多图识别，单独处理
        self.identify_images = make_dir_templates_without_number(f"{IMG_ROOT}/identify_images")

        for sub_folder in os.listdir(IMG_ROOT):
            if sub_folder not in self.__dict__:
                sub_folder_path = os.path.join(IMG_ROOT, sub_folder)
                self.__dict__.update({sub_folder: make_dir_templates(sub_folder_path)})


IMG = ImageSet()



"""获取 UI 树
"""


class SwitchMethod:
    def __init__(self, fun_list):
        self.operates = fun_list

    def operate(self):
        res = list(self.operates)
        return res

    def print(self):
        for operation in self.operates:
            print(operation, end=" ")


class Node:
    """节点
    保存 UI 树的节点
    """

    def __init__(self, name, id):
        self.id = id
        self.name = name
        self.father_edge = None
        self.father = None
        self.depth = 0
        self.edges = []

    def set_father(self, father):
        self.father = father
        self.father_edge = self.find_edge(father)

    def add_edge(self, edge):
        self.edges.append(edge)

    def find_edges(self, v):
        return [edge for edge in self.edges if edge.v is v]

    def find_edge(self, v):
        from random import choice

        edges = self.find_edges(v)
        return choice(edges) if len(edges) != 0 else None

    def print(self):
        print("节点名:", self.name, "节点编号:", self.id)
        print("父节点:", self.father)
        print("节点连边:")
        for edge in self.edges:
            edge.print()

    def __str__(self):
        return self.name


class Edge:
    """图的边
    保存 UI 图的边
    """

    def __init__(
        self,
        operate_fun: SwitchMethod,
        u: Node,
        v: Node,
        other_dst=None,
        extra_op: SwitchMethod = None,
    ):
        self.operate_fun = operate_fun
        self.u = u
        self.v = v
        self.other_dst = other_dst
        self.extra_op = extra_op

    def operate(self):
        return self.operate_fun.operate()

    def print(
        self,
    ):
        print("起点:", self.u.name, "终点:", self.v.name)


class UI:
    """UI树(拓扑学概念)"""

    def __init__(self):
        self.nodes = {}
        self.page_count = 0

        self._build_ui_tree()

    def get_node_by_name(self, name):
        return self.nodes.get(name)

    def page_exist(self, page):
        return page in self.nodes.values()

    def find_path(self, start: Node, end: Node):
        lca = self._lca(start, end)
        path1, path2 = [], []

        while start != lca:
            path1.append(start)
            start = start.father
        while end != lca:
            path2.append(end)
            end = end.father
        path2.reverse()
        path = path1 + [lca] + path2
        result, i = [], 0
        while i < len(path):
            node = path[i]
            result.append(node)
            if node.find_edge(path[-1]) != None:
                return result + [path[-1]]
            for j in range(i + 2, len(path)):
                dst = path[j]
                if node.find_edge(dst) != None:
                    i = j - 1
            i += 1
        return result

    def print(self):
        for node in self.nodes.values():
            node.print()

    def _build_ui_tree(self):
        main_page = self._construct_node("main_page", None)
        (
            map_page,
            exercise_page,
            expedition_page,
            battle_page,
            decisive_battle_entrance,
        ) = self._construct_integrative_pages(
            main_page,
            names=[
                "map_page",
                "exercise_page",
                "expedition_page",
                "battle_page",
                "decisive_battle_entrance",
            ],
            click_positions=[(163, 25), (287, 25), (417, 25), (544, 25), (661, 25)],
            common_edges=[{"pos": (30, 30), "dst": main_page}],
        )

        options_page = self._construct_node("options_page", main_page)
        (
            build_page,
            destroy_page,
            develop_page,
            discard_page,
        ) = self._construct_integrative_pages(
            options_page,
            names=[
                "build_page",
                "destroy_page",
                "develop_page",
                "discard_page",
            ],
            click_positions=[(163, 25), (287, 25), (417, 25), (544, 25)],
            common_edges=[{"pos": (30, 30), "dst": options_page}],
        )
        intensify_page, remake_page, skill_page = self._construct_integrative_pages(
            options_page,
            names=[
                "intensify_page",
                "remake_page",
                "skill_page",
            ],
            click_positions=[(163, 25), (287, 25), (417, 25)],
            common_edges=[{"pos": (30, 30), "dst": options_page}],
        )

        fight_prepare_page = self._construct_node("fight_prepare_page", map_page)
        backyard_page = self._construct_node("backyard_page", main_page)
        bath_page = self._construct_node("bath_page", backyard_page)
        choose_repair_page = self._construct_node("choose_repair_page", bath_page)
        canteen_page = self._construct_node("canteen_page", backyard_page)
        mission_page = self._construct_node("mission_page", main_page)
        support_set_page = self._construct_node("support_set_page", main_page)
        friend_page = self._construct_node("friend_page", options_page)

        self._add_edge(
            main_page,
            map_page,
            self._construct_clicks_method([(900, 480, 1, 0)]),
            expedition_page,
        )
        self._add_edge(main_page, mission_page, self._construct_clicks_method([(656, 480, 1, 0)]))
        self._add_edge(main_page, backyard_page, self._construct_clicks_method([(45, 80, 1, 0)]))
        self._add_edge(
            main_page,
            support_set_page,
            self._construct_clicks_method([(50, 135, 1, 1), (200, 300, 1, 1)]),
        )
        self._add_edge(main_page, options_page, self._construct_clicks_method([(42, 484, 1, 0)]))

        self._add_edge(
            map_page,
            fight_prepare_page,
            self._construct_clicks_method([(600, 300, 1, 0)]),
        )

        self._add_edge(
            fight_prepare_page,
            map_page,
            self._construct_clicks_method([(30, 30, 1, 0)]),
        )
        self._add_edge(
            fight_prepare_page,
            bath_page,
            self._construct_clicks_method([(840, 20, 1, 0)]),
        )

        self._add_edge(
            options_page,
            build_page,
            self._construct_clicks_method([(150, 200, 1, 1.25), (360, 200, 1, 0)]),
            develop_page,
        )
        self._add_edge(
            options_page,
            remake_page,
            self._construct_clicks_method([(150, 270, 1, 1.25), (360, 270, 1, 0)]),
        )
        self._add_edge(options_page, friend_page, self._construct_clicks_method([(150, 410, 1, 0)]))
        self._add_edge(options_page, main_page, self._construct_clicks_method([(36, 500, 1, 0)]))

        self._add_edge(
            backyard_page,
            canteen_page,
            self._construct_clicks_method([(700, 400, 1, 0)]),
        )
        self._add_edge(backyard_page, bath_page, self._construct_clicks_method([(300, 200, 1, 0)]))
        self._add_edge(backyard_page, main_page, self._construct_clicks_method([(50, 30, 1, 0)]))

        self._add_edge(bath_page, main_page, self._construct_clicks_method([(120, 30, 1, 0)]))
        self._add_edge(
            bath_page,
            choose_repair_page,
            self._construct_clicks_method([(900, 30, 1, 0)]),
        )

        self._add_edge(
            choose_repair_page,
            bath_page,
            self._construct_clicks_method([(916, 45, 1, 0)]),
        )

        self._add_edge(canteen_page, main_page, self._construct_clicks_method([(120, 30, 1, 0)]))
        self._add_edge(canteen_page, backyard_page, self._construct_clicks_method([(50, 30, 1, 0)]))

        self._add_edge(mission_page, main_page, self._construct_clicks_method([(30, 30, 1, 0)]))

        self._add_edge(
            support_set_page,
            main_page,
            self._construct_clicks_method([(30, 30, 1, 0.5), (50, 30, 1, 0.5)]),
        )

        self._add_edge(friend_page, options_page, self._construct_clicks_method([(30, 30, 1, 0)]))

        self._dfs(main_page)

    def _add_node(self, node: Node):
        self.nodes[node.name] = node

    def _lca(self, u: Node, v: Node) -> Node:
        if v.depth > u.depth:
            return self._lca(v, u)
        if u == v:
            return v
        if u.depth == v.depth:
            return self._lca(u.father, v.father)
        return self._lca(u.father, v)

    def _dfs(self, u: Node):
        for edge in u.edges:
            son = edge.v
            if son == u.father or son.father != u:
                continue

            son.depth = u.depth + 1
            self._dfs(son)

    def _list_walk_path(self, start: Node, end: Node):
        path = self.find_path(start, end)
        for node in path:
            print(node, end="->")

    def _construct_node(self, name: str, father):
        self.page_count += 1
        node = Node(name, self.page_count)
        node.set_father(father)
        self._add_node(node)
        return node

    def _construct_clicks_method(self, click_position_args):
        operations = [["click", operation] for operation in click_position_args]

        return SwitchMethod(operations)

    def _add_edge(self, u: Node, v: Node, method: SwitchMethod, other_dst=None):
        edge = Edge(method, u, v, other_dst=other_dst)
        u.add_edge(edge)

    def _construct_integrative_pages(self, father, click_positions=[], names=[], common_edges=[]):
        assert len(click_positions) == len(names)
        first_node = self._construct_node(names[0], father)
        nodes = [first_node]
        for name in names[1:]:
            nodes.append(self._construct_node(name, first_node))
        for i, node in enumerate(nodes):
            for j, click_position in enumerate(click_positions):
                if i == j:
                    continue
                self._add_edge(node, nodes[j], self._construct_clicks_method([click_position]))

            for edge in common_edges:
                dst = edge.get("dst")
                x, y = edge.get("pos")
                self._add_edge(node, dst, self._construct_clicks_method([(x, y)]))
        return nodes


WSGR_UI = UI()



from typing import Iterable


def remove_0_value_from_dict(d: dict):
    rd = {}
    for key, value in d.items():
        if value != 0 and value != None:
            rd[key] = value
    return rd


def unzip_element(o: object):
    if not isinstance(o, Iterable):
        return [o]
    res = []
    for value in o:
        if isinstance(value, (list, set)):
            res += unzip_element(value)
        else:
            res.append(value)
    return res


def unorder_equal(o1: Iterable, o2: Iterable, skip=[None]):
    for obj in o1:
        if obj not in o2 and obj not in skip:
            return False
    for obj in o2:
        if obj not in o1 and obj not in skip:
            return False
    return True



from paddleocr import PaddleOCR, draw_ocr
from AutoWSGR.utils.api_image import crop_image


en_paddle_ocr = None
cn_paddle_ocr = None

def load_en_ocr():
    global en_paddle_ocr
    en_paddle_ocr = PaddleOCR(use_angle_cls=True, lang="en",show_log = False)  # need to run only once to download and load model into memory


def load_ch_ocr():
    global cn_paddle_ocr
    cn_paddle_ocr = PaddleOCR(use_angle_cls=True, lang="ch",show_log = False)  # need to run only once to download and load model into memory

def paddle_ocr(image, language = "en"):

    if language == "ch":
        if cn_paddle_ocr == None:
            load_ch_ocr()
        result = cn_paddle_ocr.ocr(image , cls=True)
    elif language == "en" :
        if en_paddle_ocr == None:
            load_en_ocr()
        result = en_paddle_ocr.ocr(image,cls=True)
    #print(f"原始数据为： {result}")
    try:
        return result[0]
    except:
        result = None
    return result 

"""if __name__ == "__main__":
    image="4.PNG"
    #image_crop = crop_image(image, resolution=timer.config.resolution)
    result = paddle_ocr(image,language = "ch")
    print(result)"""





import numpy as np

from autowsgr.constants.image_templates import MyTemplate


def relative_to_absolute(record_pos, resolution=(960, 540)):
    """将相对坐标转换为绝对坐标"""
    delta_x, delta_y = record_pos
    _w, _h = resolution
    target_x = _w * (0.5 + delta_x)
    target_y = _h * (0.5 + delta_y)
    return target_x, target_y


def absolute_to_relative(absolute_pos, resolution=(960, 540)):
    """将绝对坐标转换为相对坐标"""
    _w, _h = resolution
    delta_x = (absolute_pos[0] - _w * 0.5) / _w
    delta_y = (absolute_pos[1] - _h * 0.5) / _h
    return delta_x, delta_y


def crop_image(image, pos1, pos2, resolution=(960, 540)):
    """按照给定的位置裁剪图片, pos1 左下角, pos2 右上角"""
    x1, y2 = map(int, relative_to_absolute(pos1, resolution))
    x2, y1 = map(int, relative_to_absolute(pos2, resolution))
    return image[y1:y2, x1:x2]


def convert_position(x, y, resolution, mode="960_to_this"):
    """转化坐标格式(放缩)

    Args:
        x (int): 横坐标
        y (int): 纵坐标
        mode (str): 工作模式
            values:
                '960_to_this':将 960x540 格式转化为当前模拟器屏幕坐标
                'this_to_960':将当前模拟器屏幕坐标转化为 960x540 格式

    Returns:Tuple(int,int)

        mode='960_to_this': 对应到当前模拟器屏幕上的坐标
        mode='this_to_960': 960x540 格式坐标
    Raise:
        ValueError:如果不支持这个模式
    """

    if mode == "960_to_this":
        return (int(x / 960 * resolution[0]), int(y / 540 * resolution[1]))
    if mode == "this_to_960":
        return (int(x * 960 / resolution[0]), int(y * 540 / resolution[1]))
    raise ValueError("unsupported mode " + str(mode))


def convert_area(area, resolution, mode="960_to_this"):
    """转化矩阵格式(放缩)

    Args:
        area list(left, top, right, button): 矩阵,列表或者元组
        mode (str): 工作模式
            values:
                '960_to_this':将 960x540 格式转化为当前模拟器屏幕坐标
                'this_to_960':将当前模拟器屏幕坐标转化为 960x540 格式

    Returns:
        (left, top, right, button): 转化后的矩阵(元组)
    """
    left, top = convert_position(area[0], area[1], resolution, mode)
    right, button = convert_position(area[2], area[3], resolution, mode)
    return (left, top, right, button)


def locateCenterOnImage(image: np.ndarray, query: MyTemplate, confidence=0.85, this_methods=None):
    """从原图像中尝试找出一个置信度相对于模板图像最高的矩阵区域的中心坐标

    Args:
        image (np.ndarray): 原图像
        query (MyTemplate): 模板图像
        confidence (float, optional): 置信度阈值. Defaults to 0.85.
        this_methods (list, optional): 匹配方式. Defaults to ['tpl'].

    Returns:
        如果匹配结果中有超过阈值的,返回置信度最高的结果的中心绝对坐标:Tuple(int,int)

        否则返回 None
    """
    if this_methods is None:
        this_methods = ["tpl"]
    query.threshold = confidence
    match_pos = query.match_in(image, this_methods=this_methods)
    return match_pos or None



import copy
import time
from abc import ABC, abstractmethod

from autowsgr.constants import literals
from autowsgr.constants.custom_exceptions import ImageNotFoundErr, NetworkErr
from autowsgr.constants.image_templates import IMG
from autowsgr.constants.other_constants import ALL_SHIP_TYPES, SAP
from autowsgr.constants.positions import BLOOD_BAR_POSITION
from autowsgr.constants.ui import Node
from autowsgr.controller.emulator import Emulator
from autowsgr.controller.run_timer import Timer
from autowsgr.game.game_operation import (
    DestroyShip,
    Expedition,
    click_result,
    detect_ship_stats,
    get_ship,
    match_night,
)
from autowsgr.game.get_game_info import get_enemy_condition
from autowsgr.utils.io import recursive_dict_update, yaml_to_dict
from autowsgr.utils.math_functions import get_nearest
from autowsgr.utils.operator import remove_0_value_from_dict


def start_march(timer: Timer, position=(900, 500)):
    timer.Android.click(*position, 1, delay=0)
    start_time = time.time()
    while timer.identify_page("fight_prepare_page"):
        if time.time() - start_time > 3:
            timer.Android.click(*position, 1, delay=0)
            time.sleep(1)
        if timer.image_exist(IMG.symbol_image[3], need_screen_shot=0):
            return literals.DOCK_FULL_FLAG
        if timer.image_exist(IMG.symbol_image[9], need_screen_shot=0, confidence=0.8):
            time.sleep(1)
            return literals.BATTLE_TIMES_EXCEED
        if time.time() - start_time > 15:
            if timer.process_bad_network():
                if timer.identify_page("fight_prepare_page"):
                    return start_march(timer, position)
                else:
                    NetworkErr("stats unknow")
            else:
                raise TimeoutError("map_fight prepare timeout")
    return literals.OPERATION_SUCCESS_FLAG


class Ship:
    """用于表示一艘船的数据结构, 注意友方与敌方所独有的field"""

    def __init__(self) -> None:
        # 友方与敌方通用
        self.name = ""  # 舰船名称
        self.ship_type = 0  # TODO: 舰船类型，用字面常量实现
        self.health = 0  # 舰船生命值
        self.ship_stats = 0  # TODO：舰船状态，主要用于记录击沉、空位

        # 友方
        self.level = 0  # 舰船等级
        self.exp = 0  # 舰船经验值
        self.friendliness = 0  # 舰船好感度


class FightResultInfo:
    def __init__(self, timer: Timer, ship_stats) -> None:
        mvp_pos = timer.get_image_position(IMG.fight_image[14])
        self.mvp = get_nearest((mvp_pos[0], mvp_pos[1] + 20), BLOOD_BAR_POSITION[1])
        self.result = timer.wait_images(IMG.fight_result, timeout=5)
        self.ship_stats = detect_ship_stats(timer, "sumup", ship_stats)
        if timer.image_exist(IMG.fight_result["SS"], need_screen_shot=False):
            self.result = "SS"
        if self.result is None:
            timer.log_screen()
            timer.logger.warning("can't identify fight result, screen logged")

    def __str__(self):
        return f"MVP 为 {self.mvp} 号位, 战果为 {self.result}"

    def __lt__(self, other):  # <
        order = ["D", "C", "B", "A", "S", "SS"]
        if isinstance(other, FightResultInfo):
            other = other.result

        order.index(self.result) < order.index(other)

    def __le__(self, other):  # <=
        order = ["D", "C", "B", "A", "S", "SS"]
        if isinstance(other, FightResultInfo):
            other = other.result

        order.index(self.result) <= order.index(other)

    def __gt__(self, other):  # >
        return not (self <= other)

    def __ge__(self, other):  # >=
        return not (self < other)


class FightEvent:
    """战斗事件类
    事件列表: 战况选择, 获取资源, 索敌成功, 迂回, 阵型选择, 进入战斗, 是否夜战, 战斗结算, 获取舰船, 继续前进, 自动回港

    状态: 一个字典, 一共三种键值
        position: 位置, 所有事件均存在

        ship_stats: 我方舰船状态(仅在 "继续前进" 事件存在)

        enemys: 敌方舰船(仅在 "索敌成功" 事件存在), 字典或 "索敌失败"

        info: 其它额外信息(仅在 "自动回港" 事件存在)

    动作: 一个字符串
        "继续": 获取资源, 迂回, 战斗结算, 获取舰船, 自动回港等不需要决策的操作

        数字字符串: 战况选择的决策

        "SL"/数字字符串: 阵型选择的决策

        "继续"/ "SL": 进入战斗后的决策

        "战斗"/"撤退"/"迂回": 索敌成功的决策

        "追击"/"撤退": 夜战的决策

        "回港/前进": 是否前进的选择(战斗结算完毕后)

    结果: 一个字符串
        "无": 战况选择, 获取资源, 索敌成功, 阵型选择, 进入战斗, 是否夜战, 继续前进, 自动回港,

        (FightResultInfo): 表示战果信息, 战果结算

        舰船名: 获取舰船
    """

    def __init__(self, event, stats, action="继续", result="无"):
        self.event = event
        self.stats = stats
        self.action = action
        self.result = result

    def __str__(self) -> str:
        return f"事件:{self.event}, 状态:{self.stats}, 动作:{self.action}, 结果:{str(self.result)}"

    def __repr__(self) -> str:
        return f"FightEvent({self.event}, {self.stats}, {self.action}, {self.result})"


class FightHistory:
    """记录并处理战斗历史信息"""

    events = []

    def __init__(self) -> None:
        pass

    def add_event(self, event, point, action="继续", result="无"):
        self.events.append(FightEvent(event, point, action, result))

    def reset(self):
        self.events = []

    def get_fight_results(self):
        results_dict = {}
        results_list = []
        for event in self.events:
            if event.event == "战果结算":
                if event.stats["position"].isalpha():
                    results_dict[event.stats["position"]] = event.result
                else:
                    results_list.append(event.result)
        return results_list if len(results_list) else results_dict

    def get_last_point(self):
        return self.events[-1].stats["position"]

    def __str__(self) -> str:
        return "".join(str(event) + "\n" for event in self.events)


class FightInfo(ABC):
    """存储战斗中需要用到的所有状态信息, 以及更新逻辑"""

    def __init__(self, timer: Timer) -> None:
        self.timer = timer
        self.config = timer.config
        self.logger = timer.logger

        self.successor_states = {}  # 战斗流程的有向图建模，在不同动作有不同后继时才记录动作
        self.state2image = {}  # 所需用到的图片模板。格式为 [模板，等待时间]
        self.after_match_delay = {}  # 匹配成功后的延时。格式为 {状态名 : 延时时间(s),}
        self.last_state = ""
        self.last_action = ""
        self.state = ""
        self.enemys = {}  # 敌方舰船列表
        self.ship_stats = []  # 我方舰船血量列表
        self.oil = 10  # 我方剩余油量
        self.ammo = 10  # 我方剩余弹药量
        self.fight_history = FightHistory()  # 战斗结果记录

    def update_state(self):
        self.last_state = self.state

        # 计算当前可能的状态
        possible_states = copy.deepcopy(self.successor_states[self.last_state])
        if isinstance(possible_states, dict):
            possible_states = possible_states[self.last_action]
        modified_timeout = [-1 for _ in possible_states]  # 某些状态需要修改等待时间
        for i, state in enumerate(possible_states):
            if isinstance(state, list):
                state, timeout = state
                possible_states[i] = state
                modified_timeout[i] = timeout
        if self.config.SHOW_MATCH_FIGHT_STAGE:
            self.logger.debug("waiting:", possible_states, "  ")
        images = [self.state2image[state][0] for state in possible_states]
        timeout = [self.state2image[state][1] for state in possible_states]
        confidence = min([0.8] + [self.state2image[state][2] for state in possible_states if len(self.state2image[state]) >= 3])
        timeout = [timeout[i] if modified_timeout[i] == -1 else modified_timeout[i] for i in range(len(timeout))]
        timeout = max(timeout)
        # 等待其中一种出现
        fun_start_time = time.time()
        while time.time() - fun_start_time <= timeout:
            self._before_match()

            # 尝试匹配
            ret = [self.timer.images_exist(image, False, confidence=confidence) for image in images]
            if any(ret):
                self.state = possible_states[ret.index(True)]
                # 查询是否有匹配后延时
                if self.state in self.after_match_delay:
                    delay = self.after_match_delay[self.state]
                    time.sleep(delay)

                if self.config.SHOW_MATCH_FIGHT_STAGE:
                    self.logger.info(f"matched: {self.state}")
                self._after_match()

                return self.state

        # 匹配不到时报错
        self.logger.error(f"匹配状态失败! state: {self.state}  last_action: {self.last_action}")
        self.timer.log_screen(True)
        for image in images:
            self.logger.log_image(image, f"match_{str(time.time())}.PNG")
        raise ImageNotFoundErr()

    def _before_match(self):
        """每一轮尝试匹配状态前执行的操作"""
        pass

    def _after_match(self):
        """匹配到状态后执行的操作"""
        if self.state == "spot_enemy_success":
            self.enemys = get_enemy_condition(self.timer, "fight")
        if self.state == "result":
            result = FightResultInfo(self.timer, self.ship_stats)
            self.ship_stats = result.ship_stats
            self.fight_history.add_event(
                "战果结算",
                {"position": self.node if "node" in self.__dict__ else f"此类战斗({type(self)})不支持节点信息"},
                result=result,
            )

    @abstractmethod
    def reset(self):
        """需要记录与初始化的战斗信息"""
        pass


class FightPlan(ABC):
    def __init__(self, timer: Timer):
        # 把 timer 引用作为内置对象，减少函数调用的时候所需传入的参数
        self.timer = timer
        self.config = timer.config
        self.logger = timer.logger
        self.fight_logs = []

    def fight(self):
        self.Info.reset()  # 初始化战斗信息
        while True:
            ret = self._make_decision()
            if ret == literals.FIGHT_CONTINUE_FLAG:
                continue
            elif ret == "need SL":
                self._SL()
                return "SL"
            elif ret == literals.FIGHT_END_FLAG:
                self.timer.set_page(self.Info.end_page)
                self.fight_logs.append(self.Info.fight_history)
                return "success"

    def run_for_times(self, times, gap=1800):
        """多次执行同一任务, 自动进行远征操作
        Args:
            times (int): 任务执行总次数

            gap (int): 强制远征检查的间隔时间
        Raise:
            RuntimeError: 战斗进行时出现错误
        Returns:
            str:
                "OK": 任务正常结束

                "dock is full" 因为船坞已满并且没有设置解装因此退出任务
        """
        assert times >= 1
        expedition = Expedition(self.timer)
        last_flag = self.run() != "SL"
        for _ in range(1, times):
            if time.time() - self.timer.last_expedition_check_time >= gap:
                expedition.run(True)
                last_flag = False
            elif isinstance(self.timer.now_page, Node) and self.timer.now_page.name == "map_page":
                expedition.run(False)
                self.timer.goto_game_page("map_page")

            fight_flag = self.run(last_flag)
            last_flag = fight_flag != "SL"

            if fight_flag not in ["SL", "success"]:
                if fight_flag == "dock is full":
                    return "dock is full"
                raise RuntimeError(f"战斗进行时出现异常, 信息为 {fight_flag}")
        return "OK"

    def run(self, same_work=False):
        """主函数，负责一次完整的战斗.
        Args:
            same_work: 如果为 True, 则说明上一次战斗和当前执行的战斗相同并且上一次战斗正常退出, 用于简化操作流程
        Returns:
            str:
                'dock is full': 船坞已满并且没有设置自动解装

                'fight end': 战斗结束标志, 一般不返回这个, 和 success 相同

                'out of times': 战斗超时

                'SL': 进行了 SL 操作

                'success': 战斗流程正常结束(到达了某个结束点或者选择了回港)

        """
        # 战斗前逻辑
        ret = self._enter_fight(same_work)

        if ret == literals.OPERATION_SUCCESS_FLAG:
            pass
        elif ret == literals.DOCK_FULL_FLAG:
            # 自动解装功能
            if self.config.dock_full_destroy:
                self.timer.Android.relative_click(0.38 - 0.5, 0.565 - 0.5)
                DestroyShip(self.timer)
                return self.run(False)
            else:
                return ret
        elif ret == literals.FIGHT_END_FLAG:
            self.timer.set_page(self.Info.end_page)
            return ret
        elif ret == literals.BATTLE_TIMES_EXCEED:
            return ret
        else:
            self.logger.error("无法进入战斗, 原因未知! 屏幕状态已记录")
            self.timer.log_screen()
            raise BaseException(str(time.time()) + "enter fight error")

        # 战斗中逻辑
        return self.fight()

    def run_for_times_condition(self, times, last_point, result="S", insist_time=900):
        """有战果要求的多次运行, 使用前务必检查参数是否有误, 防止死循环

        Args:
            times: 次数

            last_point: 最后一个点

            result: 战果要求

            insist_time: 如果大于这个时间工作量未减少则退出工作

        Returns:
            str:
                "OK": 任务顺利结束

                "dock is full": 因为船坞已满并且不允许解装所以停止
        """
        if not isinstance(result, str) or not isinstance(last_point, str):
            raise TypeError(f"last_point, result must be str,but is {type(last_point)}, {type(result)}")
        if result not in ["S", "A", "B", "C", "D", "SS"]:
            raise ValueError(f"result value {result} is illegal, it should be 'A','B','C','D','S' or 'SS'")
        if len(last_point) != 1 or ord(last_point) > ord("Z") or ord(last_point) < ord("A"):
            raise ValueError("last_point should be a uppercase within 'A' to 'Z'")
        import time

        start_time, run = time.time(), False
        while times:
            ret = self.run(run)
            run = ret != "SL"
            if run == "dock is full":
                return run

            self.logger.info("战斗信息:\n" + str(self.Info.fight_history))
            fight_results = sorted(self.Info.fight_history.get_fight_results().items())
            finish = len(fight_results) and fight_results[-1][0] == last_point and fight_results[-1][1] >= result
            if not finish:
                self.timer.logger.info("不满足预设条件, 此次战斗不计入次数")
                if time.time() - start_time > insist_time:
                    return False
            else:
                start_time, times = time.time(), times - 1
                self.timer.logger.info(f"完成了一次满足预设条件的战斗, 剩余战斗次数:{times}")
        return "OK"

    def update_state(self):
        try:
            self.Info.update_state()
            state = self.Info.state
        except ImageNotFoundErr as _:
            # 处理点击延迟或者网络波动导致的匹配失败
            self.logger.error("Image Match Failed, Processing")
            if self.timer.process_bad_network(timeout=2.5):
                pass
            if self.Info.last_state == "spot_enemy_success":
                if self.timer.image_exist(IMG.fight_image[2]):
                    self.timer.Android.click(900, 500)
            if self.Info.last_state in ["proceed", "night"] and self.timer.images_exist(IMG.fight_image[5:7]):
                if self.Info.last_action == "yes":
                    self.timer.Android.click(325, 350, times=1)
                else:
                    self.timer.Android.click(615, 350, times=1)
            self.Info.update_state()
            state = self.Info.state
        return state

    @abstractmethod
    def _enter_fight(self, same_work=False) -> str:
        pass

    @abstractmethod
    def _make_decision(self) -> str:
        pass

    # =============== 战斗中通用的操作 ===============
    def _SL(self):
        self.timer.restart()
        self.timer.go_main_page()
        self.timer.set_page("main_page")


class DecisionBlock:
    """地图上一个节点的决策模块"""

    def __init__(self, timer: Timer, args):
        self.timer = timer
        self.config = timer.config
        self.logger = timer.logger

        self.__dict__.update(args)

        # 用于根据规则设置阵型
        self.set_formation_by_rule = False
        self.formation_by_rule = 0

    def _check_rules(self, enemys: dict):
        for rule in self.enemy_rules:
            condition, act = rule
            rcondition = ""
            last = 0
            for i, ch in enumerate(condition):
                if ord(ch) > ord("Z") or ord(ch) < ord("A"):
                    if last != i:
                        if condition[last:i] in ALL_SHIP_TYPES:
                            rcondition += f"enemys.get('{condition[last:i]}', 0)"
                        else:
                            rcondition += condition[last:i]
                    rcondition += ch
                    last = i + 1

            if self.config.SHOW_ENEMY_RULES:
                self.logger.info(rcondition)
            if eval(rcondition):
                return act

    def make_decision(self, state, last_state, last_action, Info: FightInfo):
        # sourcery skip: extract-method
        """单个节点的决策"""
        enemys = Info.enemys
        if state in ["fight_period", "night_fight_period"]:
            if self.SL_when_enter_fight == True:
                Info.fight_history.add_event(
                    "进入战斗",
                    {"position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息"},
                    "SL",
                )
                return None, "need SL"
            return None, literals.FIGHT_CONTINUE_FLAG

        elif state == "spot_enemy_success":
            retreat = self.supply_ship_mode == 1 and enemys.get(SAP, 0) == 0  # 功能: 遇到补给舰则战斗，否则撤退
            can_detour = self.timer.image_exist(IMG.fight_image[13])  # 判断该点是否可以迂回
            detour = can_detour and self.detour  # 由 Node 指定是否要迂回

            # 功能, 根据敌方阵容进行选择
            act = self._check_rules(enemys=enemys)

            if act == "retreat":
                retreat = True
            elif act == "detour":
                try:
                    assert can_detour, "该点无法迂回, 但是规则中指定了迂回"
                except AssertionError:
                    raise ValueError("该点无法迂回, 但在规则中指定了迂回")
                detour = True
            elif isinstance(act, int):
                self.set_formation_by_rule = True
                self.formation_by_rule = act

            if retreat:
                self.timer.Android.click(677, 492, delay=0.2)
                Info.fight_history.add_event(
                    "索敌成功",
                    {"position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息"},
                    "撤退",
                )
                return "retreat", literals.FIGHT_END_FLAG
            elif detour:
                try:
                    image_detour = IMG.fight_image[13]
                    Emulator.click_image(self.timer, image=image_detour)
                    self.timer.logger.info("成功执行迂回操作")
                except:
                    self.timer.logger.error("未找到迂回按钮")
                    self.timer.log_screen(True)
                    raise ImageNotFoundErr("can't found image")
                # self.timer.Android.click(540, 500, delay=0.2)
                Info.fight_history.add_event(
                    "索敌成功",
                    {"position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息"},
                    "迂回",
                )
                return "detour", literals.FIGHT_CONTINUE_FLAG

            Info.fight_history.add_event(
                "索敌成功",
                {"position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息"},
                "战斗",
            )
            if self.long_missile_support == True:
                try:
                    image_missile_support = IMG.fight_image[17]
                    Emulator.click_image(self.timer, image=image_missile_support)
                    self.timer.logger.info("成功开启远程导弹支援")
                except:
                    self.timer.logger.error("未找到远程支援按钮")
            self.timer.Android.click(855, 501, delay=0.2)
            # self.timer.Android.click(380, 520, times=2, delay=0.2) # TODO: 跳过可能的开幕支援动画，实现有问题
            return "fight", literals.FIGHT_CONTINUE_FLAG
        elif state == "formation":
            spot_enemy = last_state == "spot_enemy_success"
            value = self.formation
            if spot_enemy:
                if self.SL_when_detour_fails and last_action == "detour":
                    Info.fight_history.add_event(
                        "迂回",
                        {"position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息"},
                        result="失败",
                    )
                    Info.fight_history.add_event(
                        "阵型选择",
                        {
                            "enemys": enemys,
                            "position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息",
                        },
                        action="SL",
                    )
                    return None, "need SL"

                if self.set_formation_by_rule:
                    self.logger.debug("set formation by rule:", self.formation_by_rule)
                    value = self.formation_by_rule
                    self.set_formation_by_rule = False
            else:
                if self.SL_when_spot_enemy_fails:
                    Info.fight_history.add_event(
                        "阵型选择",
                        {
                            "enemys": "索敌失败",
                            "position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息",
                        },
                        action="SL",
                    )
                    return None, "need SL"
                if self.formation_when_spot_enemy_fails != False:
                    value = self.formation_when_spot_enemy_fails
            Info.fight_history.add_event(
                "阵型选择",
                {
                    "enemys": enemys if last_state == "spot_enemy_success" else "索敌失败",
                    "position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息",
                },
                action=value,
            )
            self.timer.Android.click(573, value * 100 - 20, delay=2)
            return value, literals.FIGHT_CONTINUE_FLAG
        elif state == "night":
            is_night = self.night
            Info.fight_history.add_event(
                "是否夜战",
                {"position": Info.node if "node" in Info.__dict__ else f"此类战斗({type(Info)})不包含节点信息"},
                action="追击" if is_night else "撤退",
            )

            match_night(self.timer, is_night)
            if is_night:
                # self.timer.Android.click(325, 350)
                return "yes", literals.FIGHT_CONTINUE_FLAG
            else:
                # self.timer.Android.click(615, 350)
                return "no", literals.FIGHT_CONTINUE_FLAG

        elif state == "result":
            # time.sleep(1.5)
            # self.timer.Android.click(900, 500, times=2, delay=0.2)
            click_result(self.timer)
            return None, literals.FIGHT_CONTINUE_FLAG
        elif state == "get_ship":
            get_ship(self.timer)
            return None, literals.FIGHT_CONTINUE_FLAG
        else:
            self.logger.error("===========Unknown State==============")
            raise BaseException()


class IndependentFightPlan(FightPlan):
    def __init__(
        self,
        timer: Timer,
        end_image,
        plan_path=None,
        default_path="plans/default.yaml",
        *args,
        **kwargs,
    ):
        """创建一个独立战斗模块, 处理从形如战役点击出征到收获舰船(或战果结算)的整个过程
        Args:
            end_image (MyTemplate): 整个战斗流程结束后的图片
        """
        super().__init__(timer)
        default_args = yaml_to_dict(default_path)
        node_defaults = default_args["node_defaults"]
        node_args = yaml_to_dict(plan_path) if (plan_path is not None) else kwargs
        node_args = recursive_dict_update(node_defaults, node_args)
        self.decision_block = DecisionBlock(timer, node_args)
        self.Info = IndependentFightInfo(timer, end_image)

    def run(self):
        super().fight()

    def _make_decision(self):
        self.Info.update_state()
        if self.Info.state == "battle_page":
            return literals.FIGHT_END_FLAG

        # 进行通用NodeLevel决策
        action, fight_stage = self.decision_block.make_decision(
            self.Info.state, self.Info.last_state, self.Info.last_action, self.Info
        )
        self.Info.last_action = action
        return fight_stage


class IndependentFightInfo(FightInfo):
    def __init__(self, timer: Timer, end_image) -> None:
        super().__init__(timer)

        self.end_page = "battle_page"

        self.successor_states = {
            "proceed": ["spot_enemy_success", "formation", "fight_period"],
            "spot_enemy_success": {
                "retreat": ["battle_page"],
                "fight": ["formation", "fight_period"],
            },
            "formation": ["fight_period"],
            "fight_period": ["night", "result"],
            "night": {
                "yes": ["night_fight_period"],
                "no": [["result", 8]],
            },
            "night_fight_period": ["result"],
            "result": ["battle_page"],  # 两页战果
        }

        self.state2image = {
            "proceed": [IMG.fight_image[5], 5],
            "spot_enemy_success": [IMG.fight_image[2], 15],
            "formation": [IMG.fight_image[1], 15],
            "fight_period": [IMG.symbol_image[4], 3],
            "night": [IMG.fight_image[6], 120],
            "night_fight_period": [IMG.symbol_image[4], 3],
            "result": [IMG.fight_image[16], 60],
            "battle_page": [end_image, 5],
        }

    def reset(self):
        self.last_state = ""
        self.last_action = ""
        self.state = "proceed"

    def _before_match(self):
        # 点击加速
        if self.state in ["proceed"]:
            p = self.timer.Android.click(380, 520, delay=0, enable_subprocess=True, not_show=True)
        self.timer.update_screen()

    def _after_match(self):
        pass  # 战役的敌方信息固定，不用获取



from functools import wraps

from autowsgr.constants.global_attribute import script_end


def try_for_times(error_processor=None):
    def _try_for_times(fun):
        @wraps(fun)
        def imtemplate(*arg, **kwargs):
            if script_end == 1:
                return
            for _ in range(3):
                try:
                    return fun(*arg, **kwargs)
                except:
                    if error_processor is not None:
                        res = error_processor()
                        print("error_identified", res)
                    else:
                        pass
                return fun(*arg, **kwargs)

        return imtemplate

    return _try_for_times


def stopper(fun):
    @wraps(fun)
    def imtemplate(*arg, **kwargs):
        return "end" if (script_end == 1) else fun(*arg, **kwargs)

    return imtemplate



import copy
import os

from autowsgr.constants import literals
from autowsgr.constants.image_templates import IMG
from autowsgr.controller.run_timer import Timer
from autowsgr.game.game_operation import MoveTeam, detect_ship_stats, quick_repair
from autowsgr.game.get_game_info import get_enemy_condition, get_exercise_stats
from autowsgr.utils.io import recursive_dict_update, yaml_to_dict

from .common import DecisionBlock, FightInfo, FightPlan, Ship, start_march

"""
演习决策模块
"""


class ExerciseDecisionBlock(DecisionBlock):
    def make_decision(self, state, last_state, last_action, Info: FightInfo):
        if state == "rival_info":
            max_times = self.max_refresh_times
            self.formation_chosen = self.formation
            while max_times >= 0:
                Info.enemys = get_enemy_condition(self.timer)
                act = self._check_rules(Info.enemys)
                if act == "refresh":
                    if max_times > 0:
                        max_times -= 1
                        self.timer.Android.click(665, 400, delay=0.75)
                    else:
                        if False:
                            # if self.discard:
                            self.timer.Android.click(878, 136, delay=1)
                            return "discard", literals.FIGHT_CONTINUE_FLAG
                        else:
                            break
                elif isinstance(act, int):
                    self.formation_chosen = act
                elif act == None:
                    break

            self.timer.Android.click(804, 390, delay=0)
            return "fight", literals.FIGHT_CONTINUE_FLAG

        elif state == "fight_prepare_page":
            MoveTeam(self.timer, self.fleet_id)
            Info.ship_stats = detect_ship_stats(self.timer)
            quick_repair(self.timer, ship_stats=Info.ship_stats)
            if start_march(self.timer) != literals.OPERATION_SUCCESS_FLAG:
                return self.make_decision(state, last_state, last_action, Info)
            return None, literals.FIGHT_CONTINUE_FLAG

        elif state == "spot_enemy_success":
            self.timer.Android.click(900, 500, delay=0)
            return None, literals.FIGHT_CONTINUE_FLAG

        elif state == "formation":
            self.timer.Android.click(573, self.formation_chosen * 100 - 20, delay=2)
            return None, literals.FIGHT_CONTINUE_FLAG

        return super().make_decision(state, last_state, last_action, Info)


class NormalExerciseInfo(FightInfo):
    """存储战斗中需要用到的所有状态信息"""

    def __init__(self, timer: Timer) -> None:
        super().__init__(timer)
        self.end_page = "exercise_page"
        self.robot = True
        self.successor_states = {
            "exercise_page": ["rival_info"],
            "rival_info": {
                "fight": ["fight_prepare_page"],
                "discard": ["exercise_page"],
            },
            "fight_prepare_page": ["spot_enemy_success", "formation", "fight_period"],
            "spot_enemy_success": ["formation", "fight_period"],
            "formation": ["fight_period"],
            "fight_period": ["night", "result"],
            "night": {
                "yes": ["night_fight_period"],
                "no": [["result", 7]],
            },
            "night_fight_period": ["result"],
            "result": ["exercise_page"],  # 两页战果
        }

        self.state2image = {
            "exercise_page": [IMG.identify_images["exercise_page"], 7.5],
            "rival_info": [IMG.exercise_image["rival_info"], 7.5],
            "fight_prepare_page": [IMG.identify_images["fight_prepare_page"], 7.5],
            "spot_enemy_success": [IMG.fight_image[2], 15],
            "formation": [IMG.fight_image[1], 15],
            "fight_period": [IMG.symbol_image[4], 10],
            "night": [IMG.fight_image[6], 0.85, 180],
            "night_fight_period": [IMG.symbol_image[4], 10],
            "result": [IMG.fight_image[3], 90],
        }

        self.after_match_delay = {
            "night": 1.75,
        }

    def reset(self):
        self.fight_history.reset()
        self.state = "rival_info"  # 初始状态等同于 "rival_info" 选择 'discard'
        self.last_action = "discard"
        # TODO: 舰船信息，暂时不用
        self.ally_ships = [Ship() for _ in range(6)]  # 我方舰船状态
        self.enemy_ships = [Ship() for _ in range(6)]  # 敌方舰船状态

    def _before_match(self):
        # 点击加速
        if self.state in ["fight_prepare_page"]:
            p = self.timer.Android.click(380, 520, delay=0, enable_subprocess=True, not_show=True)

        self.timer.update_screen()

    @property
    def node(self):
        return self.timer.ship_point


class NormalExercisePlan(FightPlan):
    """ " 常规战斗的决策模块"""

    def __init__(self, timer: Timer, plan_path):
        super().__init__(timer)

        # 加载默认配置
        default_args = yaml_to_dict(os.path.join(self.config.PLAN_ROOT, "default.yaml"))
        plan_defaults = default_args["exercise_defaults"]
        plan_defaults.update({"node_defaults": default_args["node_defaults"]})

        # 加载计划配置
        plan_args = yaml_to_dict(os.path.join(self.config.PLAN_ROOT, plan_path))
        args = recursive_dict_update(plan_defaults, plan_args, skip=["node_args"])
        self.__dict__.update(args)

        # 加载节点配置
        node_defaults = self.node_defaults
        self.nodes = {}
        for node_name in self.selected_nodes:
            node_args = copy.deepcopy(node_defaults)
            if node_name in plan_args["node_args"]:
                node_args = recursive_dict_update(node_args, plan_args["node_args"][node_name])
            self.nodes[node_name] = ExerciseDecisionBlock(timer, node_args)

        # 构建信息存储结构
        self.Info = NormalExerciseInfo(self.timer)

    def _enter_fight(self, same_work=False) -> str:
        """
        从任意界面进入战斗.

        :return: 进入战斗状态信息，包括['success', 'dock is full'].
        """
        if same_work == False:
            self.timer.goto_game_page("exercise_page")

        self._exercise_times = self.exercise_times
        self.exercise_stats = [None, None]
        return literals.OPERATION_SUCCESS_FLAG

    def _make_decision(self):
        self.Info.update_state()
        state = self.Info.state
        # 进行MapLevel的决策
        if state == "exercise_page":
            self.exercise_stats = get_exercise_stats(self.timer, self.exercise_stats[1])
            if self._exercise_times > 0 and any(self.exercise_stats[2:]):
                pos = self.exercise_stats[2:].index(True)
                self.rival = "player"
                self.timer.Android.click(770, (pos + 1) * 110 - 10)
                return literals.FIGHT_CONTINUE_FLAG
            elif self.robot and self.exercise_stats[1]:
                self.timer.Android.swipe(800, 200, 800, 400)  # 上滑
                self.timer.Android.click(770, 100)
                self.rival = "robot"
                self.exercise_stats[1] = False
                return literals.FIGHT_CONTINUE_FLAG

            else:
                return literals.FIGHT_END_FLAG

        # 进行通用NodeLevel决策
        action, fight_stage = self.nodes[self.rival].make_decision(
            state, self.Info.last_state, self.Info.last_action, self.Info
        )
        self.Info.last_action = action
        return fight_stage



import os
import subprocess
import time
from subprocess import check_output

import airtest.core.android.android
from airtest.core.api import connect_device

from autowsgr.constants.custom_exceptions import CriticalErr
from autowsgr.constants.data_roots import ADB_ROOT
from autowsgr.utils.function_wrapper import try_for_times

# Win 和 Android 的通信
# Win 向系统写入数据


class WindowsController:
    def __init__(self, config, logger) -> None:
        self.logger = logger

        self.emulator = config["type"]  # 模拟器类型
        self.emulator_name = config["emulator_name"]
        self.start_cmd = config["start_cmd"]  # 模拟器启动命令
        if self.emulator == "蓝叠 Hyper-V":
            assert config["config_file"], "Bluestacks 需要提供配置文件"
            self.emulator_config_file = config["config_file"]

        self.exe_name = os.path.basename(self.start_cmd)  # 自动获得模拟器的进程名

    # ======================== 网络 ========================
    def check_network(self):
        """检查网络状况

        Returns:
            bool:网络正常返回 True,否则返回 False
        """
        return os.system("ping www.moefantasy.com") == 0

    def wait_network(self, timeout=1000):
        """等待到网络恢复"""
        start_time = time.time()
        while time.time() - start_time <= timeout:
            if self.check_network():
                return True
            time.sleep(10)

        return False

    # ======================== 模拟器 ========================

    # @try_for_times()
    def connect_android(self) -> airtest.core.android.android.Android:
        """连接指定安卓设备
        Returns:
            dev: airtest.
        """
        if not self.is_android_online():
            self.restart_android()
            time.sleep(15)

        if self.emulator == "雷电":
            dev_name = f"ANDROID:///{self.emulator_name}"
        elif self.emulator == "蓝叠 Hyper-V":
            with open(self.emulator_config_file, "r") as f:
                lines = f.readlines()
            for line in lines:
                if line.startswith("bst.instance.Pie64.status.adb_port="):
                    port = line.split("=")[-1].strip()[1:-1]
                    dev_name = f"ANDROID:///127.0.0.1:{port}"

        from logging import ERROR, getLogger

        getLogger("airtest").setLevel(ERROR)

        start_time = time.time()
        while time.time() - start_time <= 30:
            try:
                dev = connect_device(dev_name)
                self.logger.info("Android Connected!")
                return dev
            except:
                pass

        self.logger.error("连接模拟器失败！")
        raise CriticalErr("连接模拟器失败！")

    def is_android_online(self):
        """判断 timer 给定的设备是否在线
        Returns:
            bool: 在线返回 True, 否则返回 False
        """
        raw_res = check_output(f'tasklist /fi "ImageName eq {self.exe_name}').decode("gbk")  # TODO: 检查是否所有windows版本返回都是中文
        return "PID" in raw_res

    def kill_android(self):
        try:
            subprocess.run(["taskkill", "-f", "-im", self.exe_name])
        except:
            pass

    def start_android(self):
        try:
            os.popen(self.start_cmd)
            start_time = time.time()
            while not self.is_android_online():
                time.sleep(1)
                if time.time() - start_time > 120:
                    raise TimeoutError("模拟器启动超时！")
        except Exception as e:
            self.logger.error(f"{e} 请检查模拟器路径!")
            raise CriticalErr("on Restart Android")

    def restart_android(self):
        self.kill_android()
        self.start_android()



import os
import threading as th
import time

from autowsgr.constants.custom_exceptions import (
    CriticalErr,
    ImageNotFoundErr,
    NetworkErr,
)
from autowsgr.constants.data_roots import DATA_ROOT
from autowsgr.constants.image_templates import IMG
from autowsgr.constants.other_constants import ALL_PAGES, NO
from autowsgr.constants.ui import WSGR_UI, Node
from autowsgr.utils.io import yaml_to_dict
from autowsgr.utils.operator import unzip_element

from .emulator import Emulator


def try_to_get_expedition(timer: Emulator):
    timer.logger.info("Getting Expedition Rewards....")
    get, pos = False, timer.wait_image(IMG.game_ui[6], timeout=2)
    while pos:
        timer.Android.click(pos[0], pos[1], delay=1)
        timer.wait_image(IMG.fight_image[3], after_get_delay=0.25)
        timer.Android.click(900, 500, delay=1)
        timer.ConfirmOperation(must_confirm=1, delay=0.5, confidence=0.9)
        pos, get = timer.wait_image(IMG.game_ui[6], timeout=2), True
    return get


class Timer(Emulator):
    """程序运行记录器, 用于记录和传递部分数据, 同时用于区分多开, WSGR 专用"""

    # 战舰少女R专用控制器
    everyday_check = True
    ui = WSGR_UI
    ship_stats = [0, 0, 0, 0, 0, 0, 0]  # 我方舰船状态
    enemy_type_count = {}  # 字典,每种敌人舰船分别有多少
    now_page = None  # 当前所在 UI 名
    resources = None  # 当前四项资源量
    """
    以上时能用到的
    以下是暂时用不到的
    """
    last_mission_completed = 0
    got_ship_num = 0  # 当天已掉落的船
    got_loot_num = 0  # 当天已掉落的胖次
    got_ship_name = None
    ship_type = None
    ship_star = None
    last_expedition_check_time = time.time()

    def __init__(self, config, logger):
        super().__init__(config, logger)

        if not self.config.PLAN_ROOT:
            self.logger.warning(f"No PLAN_ROOT specified, default value {os.path.join(DATA_ROOT, 'plans')} will be used")
            self.config.PLAN_ROOT = os.path.join(DATA_ROOT, "plans")
        try:
            self.ship_names = unzip_element(list(yaml_to_dict(config.SHIP_NAME_PATH).values()))
        except:
            self.logger.warning(
                f"Failed to load ship_name file:{config.SHIP_NAME_PATH}\nDefault shipname file {os.path.join(DATA_ROOT,  'default_ship_names.yaml')} will be used"
            )
            ship_name_path = os.path.join(DATA_ROOT, "default_ship_names.yaml")
            self.ship_names = unzip_element(list(yaml_to_dict(ship_name_path).values()))

        self.init()

    # ========================= 初级游戏控制 =========================
    def init(self):
        """初始化游戏状态, 以便进一步的控制"""
        self.app_name = self.which_app
        # ========== 启动游戏 ==========
        if self.config.account is not None and self.config.password != None:
            self.restart(account=self.config.account, password=self.config.password)
        if self.Android.is_game_running() == False:
            self.start_game()
        self.Android.start_app(self.app_name)

        # ========== 检查游戏页面状态 ============

        try:
            self.set_page()
            self.logger.info(f"启动成功, 当前位置: {self.now_page.name}")
        except:
            if "CHECK_PAGE" in self.config.__dict__ and self.config.CHECK_PAGE:
                self.logger.warning("无法确定当前页面, 尝试重启游戏")
                self.restart()
                self.set_page()
            else:
                self.logger.warning("在无法确定页面的情况下继续.")

    def log_in(self, account, password):
        pass

    def log_out(self, account, password):
        """在登录界面登出账号"""
        pass

    def start_game(self, account=None, password=None, delay=1.0):
        """启动游戏"""
        self.Android.start_app(self.app_name)
        res = self.wait_images(
            [IMG.start_image[2]] + IMG.confirm_image[1:],
            0.85,
            gap=1,
            timeout=70 * delay,
        )

        if res is None:
            raise TimeoutError("start_app timeout")
        if res != 0:
            self.ConfirmOperation()
            if self.wait_image(IMG.start_image[2], timeout=200) == False:
                raise TimeoutError("resource downloading timeout")

        # ========= 登录 =========

        if account != None and password != None:
            self.Android.click(75, 450)
            if self.wait_image(IMG.start_image[3]) == False:
                raise TimeoutError("can't enter account manage page")
            self.Android.click(460, 380)
            if self.wait_image(IMG.start_image[4]) == False:
                raise TimeoutError("can't logout successfully")
            self.Android.click(540, 180)
            for _ in range(20):
                p = th.Thread(target=lambda: self.Android.ShellCmd("input keyevent 67"))
                p.start()
            p.join()
            self.Android.text(str(account))
            self.Android.click(540, 260)
            for _ in range(20):
                p = th.Thread(target=lambda: self.Android.ShellCmd("input keyevent 67"))
                p.start()
            p.join()
            time.sleep(0.5)
            self.Android.text(str(password))
            self.Android.click(400, 330)
            res = self.wait_images([IMG.start_image[5], IMG.start_image[2]])
            if res is None:
                raise TimeoutError("login timeout")
            if res == 0:
                raise BaseException("password or account is wrong")

        # =========== 开始游戏 ===========

        delay = 15
        while self.image_exist(IMG.start_image[2]):
            self.click_image(IMG.start_image[2])
            if self.wait_image(IMG.game_ui[3], timeout=delay, after_get_delay=0.25):
                break
            delay *= 2
            if delay > 30:
                raise ImageNotFoundErr("can't start game")
        # ============ 关闭新闻, 领取每日奖励 =============

        try:
            if self.everyday_check:
                self.logger.info("正在尝试关闭新闻, 领取奖励")
                if self.wait_image(IMG.start_image[6], timeout=2) != False:  # 新闻与公告,设为今日不再显示
                    if not self.check_pixel((70, 485), (201, 129, 54)):
                        self.Android.click(70, 485)
                    self.Android.click(30, 30)
                if self.wait_image(IMG.start_image[7], timeout=7) != False:  # 每日签到
                    self.Android.click(474, 357)
                    self.ConfirmOperation(must_confirm=1, timeout=2)
                self.everyday_check = False
            self.go_main_page()
            self.logger.info("游戏启动成功!")
        except:
            raise BaseException("fail to start game")

    def restart(self, times=0, *args, **kwargs):
        try:
            self.Android.ShellCmd(f"am force-stop {self.app_name}")
            self.Android.ShellCmd("input keyevent 3")
            self.start_game(**kwargs)
        except:
            if self.Windows.is_android_online() == False:
                pass

            elif times == 1:
                raise CriticalErr("on restart,")

            elif self.Windows.check_network() == False:
                for i in range(11):
                    time.sleep(10)
                    if self.Windows.check_network() == True:
                        break
                    if i == 10:
                        raise NetworkErr()

            elif self.Android.is_game_running():
                raise CriticalErr("CriticalErr on restart function")

            self.Windows.connect_android()
            self.restart(times + 1, *args, **kwargs)

    def is_bad_network(self, timeout=10):
        return self.wait_images([IMG.error_image["bad_network"], IMG.symbol_image[10]], timeout=timeout) != None

    def process_bad_network(self, extra_info="", timeout=10):
        """判断并处理网络状况问题
        Returns:
            bool: 如果为 True 则表示为网络状况问题,并已经成功处理,否则表示并非网络问题或者处理超时.
        Raise:
            TimeoutError:处理超时
        """
        start_time = time.time()
        while self.is_bad_network(timeout):
            self.logger.error(f"bad network: {extra_info}")
            while True:
                if time.time() - start_time >= 180:
                    raise TimeoutError("Process bad network timeout")
                if self.Windows.check_network() != False:
                    break

            start_time2 = time.time()
            while self.image_exist([IMG.symbol_image[10]] + [IMG.error_image["bad_network"]]):
                time.sleep(0.5)
                if time.time() - start_time2 >= 60:
                    break
                if self.image_exist(IMG.error_image["bad_network"]):
                    self.Android.click(476, 298, delay=2)

            if time.time() - start_time2 < 60:
                self.logger.debug("ok network problem solved, at", time.time())
                return True

        return False

    # ========================= 维护当前所在游戏界面 =========================
    def _integrative_page_identify(self):
        positions = [(171, 47), (300, 47), (393, 47), (504, 47), (659, 47)]
        for i, position in enumerate(positions):
            if self.check_pixel(position, (225, 130, 16)):
                return i + 1

    def identify_page(self, name, need_screen_shot=True):
        if need_screen_shot:
            self.update_screen()

        if (name == "main_page") and (self.identify_page("options_page", 0)):
            return False
        if (name == "map_page") and (self._integrative_page_identify() != 1 or self.check_pixel((35, 297), (47, 253, 226))):
            return False
        if (name == "build_page") and (self._integrative_page_identify() != 1):
            return False
        if (name == "develop_page") and (self._integrative_page_identify() != 3):
            return False

        return any(self.image_exist(template, 0) for template in IMG.identify_images[name])

    def wait_pages(self, names, timeout=10, gap=0.1, after_wait=0.1):
        start_time = time.time()
        if isinstance(names, str):
            names = [names]
        while True:
            self.update_screen()
            for i, name in enumerate(names):
                if self.identify_page(name, 0):
                    time.sleep(after_wait)
                    return i + 1

            if time.time() - start_time > timeout:
                break
            time.sleep(gap)

        raise TimeoutError(f"identify timeout of{str(names)}")

    def get_now_page(self):
        """获取并返回当前页面名称"""
        self.update_screen()
        for page in ALL_PAGES:
            if self.identify_page(page, need_screen_shot=False):
                return page
        return "unknown_page"

    def check_now_page(self):
        return self.identify_page(name=self.now_page.name)

    def operate(self, end: Node):
        ui_list = self.ui.find_path(self.now_page, end)
        for next in ui_list[1:]:
            edge = self.now_page.find_edge(next)
            opers = edge.operate()
            self.now_page = next
            for oper in opers:
                fun, args = oper
                if fun == "click":
                    self.Android.click(*args)
                else:
                    self.logger.error(f"unknown function name: {fun}")
                    raise BaseException()

            if edge.other_dst is not None:
                dst = self.wait_pages(names=[self.now_page.name, edge.other_dst.name])
                if dst == 1:
                    continue
                self.logger.debug(f"Go page: {self.now_page.name}, but arrive: {edge.other_dst.name}")
                self.now_page = self.ui.get_node_by_name([self.now_page.name, edge.other_dst.name][dst - 1])
                self.logger.debug(f"Now page: {self.now_page.name}")
                if self.now_page.name == "expedition_page":
                    try_to_get_expedition(self)
                self.operate(end)
                return
            else:
                self.wait_pages(names=[self.now_page.name])
            time.sleep(0.25)

    def set_page(self, page_name=None, page=None):
        if page_name is None and page is None:
            now_page = self.get_now_page()

            if now_page is None:
                raise ImageNotFoundErr("无法识别该页面")
            else:
                if now_page != "unknown_page":
                    self.now_page = self.ui.get_node_by_name(now_page)
                else:
                    self.now_page = now_page
        elif page is not None:
            if not isinstance(page, Node):
                self.logger.error("arg:page must be an controller.ui.Node object")
                raise ValueError()

            self.now_page = page if (self.ui.page_exist(page)) else "unknown_page"
        else:
            page = self.ui.get_node_by_name(page_name)
            if page is None:
                page = "unknown_page"

            self.now_page = page

    def walk_to(self, end, try_times=0):
        try:
            if isinstance(self.now_page, str) and "unknow" in self.now_page:
                self.go_main_page()
            if isinstance(end, Node):
                self.operate(end)
                self.wait_pages(end.name)
                return
            if isinstance(end, str):
                end = self.ui.get_node_by_name(end)
                if end is None:
                    self.logger.error("unacceptable value of end: {end}")
                    raise ValueError("illegal value:end, in Timer.walk_to")
                self.walk_to(end)

        except TimeoutError as exception:
            if try_times > 3:
                raise TimeoutError("can't access the page")
            if self.is_bad_network(timeout=0) == False:
                self.logger.debug("wrong path is operated,anyway we find a way to solve,processing")
                self.logger.debug("wrong info is:", exception)
                self.go_main_page()
                self.walk_to(end, try_times + 1)
            else:
                while True:
                    if self.process_bad_network("can't walk to the position because a TimeoutError"):
                        try:
                            if not self.wait_pages(names=self.now_page.name, timeout=1):
                                self.set_page(self.get_now_page())
                        except:
                            self.go_main_page()
                        else:
                            break
                    else:
                        raise ValueError("unknown error")
                self.walk_to(end)

    def go_main_page(self, QuitOperationTime=0, List=[], ExList=[]):
        """回退到游戏主页
        Args:
            timer (Timer): _description_
            QuitOperationTime (int, optional): _description_. Defaults to 0.
            List (list, optional): _description_. Defaults to [].
            ExList (list, optional): _description_. Defaults to [].

        Raises:
            ValueError: _description_
        """
        if QuitOperationTime > 200:
            raise ValueError("Error,Couldn't go main page")

        self.now_page = self.ui.get_node_by_name("main_page")
        if len(List) == 0:
            List = IMG.back_buttons[1:] + ExList
        type = self.wait_images(List + [IMG.game_ui[3]], 0.8, timeout=0)

        if type is None:
            self.go_main_page(QuitOperationTime + 1, List)
            return

        if type >= len(List):
            type = self.wait_images(List, timeout=0)
            if type is None:
                return

        pos = self.get_image_position(List[type], 0, 0.8)
        self.Android.click(pos[0], pos[1])

        NewList = List[1:] + [List[0]]
        self.go_main_page(QuitOperationTime + 1, NewList)

    def goto_game_page(self, target="main", extra_check=False):
        """到某一个游戏界面

        Args:
            target (str, str): 目标章节名(见 ./constants/other_constants). Defaults to 'main'.
        """
        self.walk_to(target)
        if extra_check:
            self.wait_pages(names=[self.now_page.name])

    def ConfirmOperation(self, must_confirm=0, delay=0.5, confidence=0.9, timeout=0):
        """等待并点击弹出在屏幕中央的各种确认按钮

        Args:
            must_confirm (int, optional): 是否必须按. Defaults to 0.
            delay (float, optional): 点击后延时(秒). Defaults to 0.5.
            timeout (int, optional): 等待延时(秒),负数或 0 不等待. Defaults to 0.

        Raises:
            ImageNotFoundErr: 如果 must_confirm = True 但是 timeout 之内没找到确认按钮排除该异常
        Returns:
            bool:True 为成功,False 为失败
        """
        pos = self.wait_images(IMG.confirm_image[1:], confidence=confidence, timeout=timeout)
        if pos is None:
            if must_confirm == 1:
                raise ImageNotFoundErr("no confirm image found")
            else:
                return False
        res = self.get_image_position(IMG.confirm_image[pos + 1], confidence=confidence, need_screen_shot=0)
        self.Android.click(res[0], res[1], delay=delay)
        return True
    @property
    def which_app(self):
        if self.config.game_app == "官服":
            start_game_app = "com.huanmeng.zhanjian2"
        elif self.config.game_app == "应用宝":
            start_game_app = "com.tencent.tmgp.zhanjian2"
        elif self.config.game_app == "小米":
            start_game_app = "com.hoolai.zjsnr.mi"
        return start_game_app


def process_error(timer: Timer):
    print("processing errors")
    if timer.Windows.is_android_online() == False or timer.Android.is_game_running() == False:
        timer.Windows.restart_android()
        timer.Windows.connect_android()

        return "Android Restarted"

    return "ok,bad network" if timer.process_bad_network() else "ok,unknown error"
